import json
from functools import partial
from unittest.mock import AsyncMock, MagicMock, call, patch

import pytest
from bluesky.run_engine import RunEngine
from bluesky.simulators import assert_message_and_return_remaining
from bluesky.utils import MsgGenerator
from dodal.beamlines import i04
from dodal.devices.beamlines.i04.murko_results import MurkoResultsDevice
from dodal.devices.oav.oav_detector import OAV, OAVBeamCentrePV, OAVConfig
from dodal.devices.oav.oav_to_redis_forwarder import OAVToRedisForwarder, Source
from dodal.devices.robot import BartRobot
from dodal.devices.smargon import Smargon
from dodal.devices.thawer import OnOff, Thawer
from ophyd_async.core import (
    completed_status,
    get_mock_put,
    init_devices,
    set_mock_value,
)

from mx_bluesky.beamlines.i04.callbacks.murko_callback import (
    FORWARDING_COMPLETE_MESSAGE,
    MurkoCallback,
)
from mx_bluesky.beamlines.i04.thawing_plan import (
    thaw,
    thaw_and_murko_centre,
    thaw_and_stream_to_redis,
)

DISPLAY_CONFIGURATION = "tests/test_data/test_display.configuration"
ZOOM_LEVELS_XML = "tests/test_data/test_jCameraManZoomLevels.xml"


class MyError(Exception):
    pass


@pytest.fixture
async def oav_full_screen() -> OAV:
    oav_config = OAVConfig(ZOOM_LEVELS_XML)
    async with init_devices(mock=True, connect=True):
        oav = OAVBeamCentrePV(
            "", config=oav_config, name="oav_full_screen", mjpeg_prefix="XTAL"
        )
    set_mock_value(oav.zoom_controller.level, "1.0x")
    set_mock_value(oav.grid_snapshot.x_size, 1024)
    set_mock_value(oav.grid_snapshot.y_size, 768)
    return oav


@pytest.fixture
async def oav_roi() -> OAV:
    oav_config = OAVConfig(ZOOM_LEVELS_XML)
    async with init_devices(mock=True, connect=True):
        oav = OAVBeamCentrePV("", config=oav_config, name="oav")
    set_mock_value(oav.zoom_controller.level, "5.0x")
    set_mock_value(oav.grid_snapshot.x_size, 512)
    set_mock_value(oav.grid_snapshot.y_size, 384)

    return oav


@pytest.fixture
async def smargon() -> Smargon:
    smargon = Smargon(prefix="BL04I-MO-SGON-01:", name="gonio")
    await smargon.connect(mock=True)

    set_mock_value(smargon.omega.user_readback, 0.0)

    return smargon


@pytest.fixture
def thawer() -> Thawer:
    return i04.thawer.build(connect_immediately=True, mock=True)


@pytest.fixture
@patch("dodal.devices.beamlines.i04.murko_results.StrictRedis")
async def murko_results(mock_strict_redis: MagicMock) -> MurkoResultsDevice:
    murko_results = MurkoResultsDevice(name="murko_results")
    murko_results.trigger = MagicMock(side_effect=completed_status)
    murko_results.stage = MagicMock(side_effect=completed_status)
    murko_results.unstage = MagicMock(side_effect=completed_status)
    return murko_results


@pytest.fixture
async def oav_forwarder(oav_full_screen: OAV, oav_roi: OAV) -> OAVToRedisForwarder:
    with init_devices(mock=True):
        oav_forwarder = OAVToRedisForwarder(
            "prefix",
            oav_roi,
            oav_full_screen,
            "host",
            "password",
            name="oav_to_redis_forwarder",
        )
    set_mock_value(oav_forwarder.uuid, "test")

    oav_forwarder.kickoff = MagicMock(side_effect=completed_status)
    oav_forwarder.complete = MagicMock(side_effect=completed_status)
    return oav_forwarder


@pytest.fixture
def robot() -> BartRobot:
    return i04.robot.build(connect_immediately=True, mock=True)


def _do_thaw_and_confirm_cleanup(
    move_mock: MagicMock, smargon: Smargon, thawer: Thawer, do_thaw_func
):
    smargon.omega.set = move_mock
    set_mock_value(smargon.omega.velocity, initial_velocity := 10)
    smargon.omega.set = move_mock
    do_thaw_func()
    last_thawer_call = get_mock_put(thawer._control).call_args_list[-1]
    assert last_thawer_call == call(OnOff.OFF)
    last_velocity_call = get_mock_put(smargon.omega.velocity).call_args_list[-1]
    assert last_velocity_call == call(initial_velocity)


def test_given_thaw_succeeds_then_velocity_restored_and_thawer_turned_off(
    smargon: Smargon, thawer: Thawer, run_engine: RunEngine
):
    def do_thaw_func():
        run_engine(thaw(10, thawer=thawer, smargon=smargon))

    _do_thaw_and_confirm_cleanup(
        MagicMock(side_effect=lambda _: completed_status()),
        smargon,
        thawer,
        do_thaw_func,
    )


def test_given_moving_smargon_gives_error_then_velocity_restored_and_thawer_turned_off(
    smargon: Smargon, thawer: Thawer, run_engine: RunEngine
):
    def do_thaw_func():
        with pytest.raises(MyError):
            run_engine(thaw(10, thawer=thawer, smargon=smargon))

    _do_thaw_and_confirm_cleanup(
        MagicMock(side_effect=MyError()), smargon, thawer, do_thaw_func
    )


@pytest.mark.parametrize(
    "time, rotation, expected_speed",
    [
        (10, 360, 72),
        (3.5, 100, pytest.approx(57.142857)),
        (50, -100, 4),
    ],
)
def test_given_different_rotations_and_times_then_velocity_correct(
    smargon: Smargon,
    thawer: Thawer,
    time: float,
    rotation: float,
    expected_speed: float,
    run_engine: RunEngine,
):
    run_engine(thaw(time, rotation, thawer=thawer, smargon=smargon))
    first_velocity_call = get_mock_put(smargon.omega.velocity).call_args_list[0]
    assert first_velocity_call == call(expected_speed)


@pytest.mark.parametrize(
    "start_pos, rotation, expected_end",
    [
        (0, 360, 360),
        (78, 100, 178),
        (0, -100, -100),
    ],
)
def test_given_different_rotations_then_motor_moved_relative(
    smargon: Smargon,
    thawer: Thawer,
    start_pos: float,
    rotation: float,
    expected_end: float,
    run_engine: RunEngine,
):
    set_mock_value(smargon.omega.user_setpoint, start_pos)
    run_engine(thaw(10, rotation, thawer=thawer, smargon=smargon))
    assert get_mock_put(smargon.omega.user_setpoint).call_args_list == [
        call(expected_end),
        call(start_pos),
    ]


@patch("mx_bluesky.beamlines.i04.thawing_plan.MurkoCallback")
async def test_thaw_and_stream_sets_sample_id_and_kicks_off_forwarder(
    patch_murko_callback: MagicMock,
    smargon: Smargon,
    thawer: Thawer,
    oav_forwarder: OAVToRedisForwarder,
    robot: BartRobot,
    run_engine: RunEngine,
):
    set_mock_value(robot.sample_id, 100)
    run_engine(
        thaw_and_stream_to_redis(
            10,
            360,
            thawer=thawer,
            smargon=smargon,
            robot=robot,
            oav_to_redis_forwarder=oav_forwarder,
        )
    )
    assert await oav_forwarder.sample_id.get_value() == 100
    oav_forwarder.kickoff.assert_called()  # type: ignore
    oav_forwarder.complete.assert_called()  # type: ignore


@patch("mx_bluesky.beamlines.i04.thawing_plan.MurkoCallback")
def test_thaw_and_stream_adds_murko_callback_and_produces_expected_messages(
    patch_murko_callback: MagicMock,
    smargon: Smargon,
    thawer: Thawer,
    robot: BartRobot,
    oav_forwarder: OAVToRedisForwarder,
    run_engine: RunEngine,
):
    patch_murko_instance = patch_murko_callback.return_value
    run_engine(
        thaw_and_stream_to_redis(
            10,
            360,
            thawer=thawer,
            smargon=smargon,
            robot=robot,
            oav_to_redis_forwarder=oav_forwarder,
        )
    )

    docs = patch_murko_instance.call_args_list
    start_params = [c.args[1] for c in docs if c.args[0] == "start"]
    event_params = [c.args[1] for c in docs if c.args[0] == "event"]
    assert len(start_params) == 2
    assert len(event_params) == 8
    oav_updates = [
        e for e in event_params if "oav_to_redis_forwarder-uuid" in e["data"].keys()
    ]
    smargon_updates = [e for e in event_params if "gonio-omega" in e["data"].keys()]
    assert len(oav_updates) > 0
    assert len(smargon_updates) > 0


@patch("mx_bluesky.beamlines.i04.thawing_plan.MurkoCallback.stop")
@patch("mx_bluesky.beamlines.i04.thawing_plan.MurkoCallback.call_murko")
@patch("mx_bluesky.beamlines.i04.thawing_plan.MurkoCallback._check_redis_connection")
def test_thaw_and_stream_will_produce_events_that_call_murko(
    patch_check_redis_connection: MagicMock,
    patch_murko_call: MagicMock,
    patch_stop_call: MagicMock,
    smargon: Smargon,
    thawer: Thawer,
    robot: BartRobot,
    oav_forwarder: OAVToRedisForwarder,
    run_engine: RunEngine,
):
    patch_check_redis_connection.return_value = True

    class StopPlanError(Exception):
        pass

    patch_stop_call.side_effect = StopPlanError()

    with pytest.raises(StopPlanError):
        run_engine(
            thaw_and_stream_to_redis(
                10,
                360,
                thawer=thawer,
                smargon=smargon,
                robot=robot,
                oav_to_redis_forwarder=oav_forwarder,
            )
        )

    patch_murko_call.assert_called()


def _test_plan_will_switch_murko_source_half_way_through_thaw(
    sim_run_engine, plan: MsgGenerator
):
    msgs = sim_run_engine.simulate_plan(plan)
    for source in [Source.FULL_SCREEN.value, Source.ROI.value]:
        msgs = assert_message_and_return_remaining(
            msgs,
            lambda msg: msg.command == "set"
            and msg.obj.name == "oav_to_redis_forwarder-selected_source"
            and msg.args[0] == source,
        )
        msgs = assert_message_and_return_remaining(
            msgs,
            lambda msg: msg.command == "kickoff"
            and msg.obj.name == "oav_to_redis_forwarder",
        )
        msgs = assert_message_and_return_remaining(
            msgs,
            lambda msg: msg.command == "set" and msg.obj.name == "gonio-omega",
        )
        msgs = assert_message_and_return_remaining(
            msgs,
            lambda msg: msg.command == "complete"
            and msg.obj.name == "oav_to_redis_forwarder",
        )


def test_thaw_and_stream_will_switch_murko_source_half_way_through_thaw(
    sim_run_engine,
    smargon: Smargon,
    thawer: Thawer,
    oav_forwarder: OAVToRedisForwarder,
    robot: BartRobot,
):
    _test_plan_will_switch_murko_source_half_way_through_thaw(
        sim_run_engine,
        thaw_and_stream_to_redis(10, 360, robot, thawer, smargon, oav_forwarder),
    )


def _run_thaw_and_stream_and_assert_zoom_changes(
    smargon: Smargon,
    thawer: Thawer,
    oav_forwarder: OAVToRedisForwarder,
    oav_full_screen: OAV,
    robot: BartRobot,
    run_engine: RunEngine,
    expect_raises=None,
):
    set_mock_value(oav_full_screen.zoom_controller.level, "2.0x")

    run_plan = partial(
        run_engine,
        thaw_and_stream_to_redis(
            10,
            360,
            thawer=thawer,
            smargon=smargon,
            robot=robot,
            oav_to_redis_forwarder=oav_forwarder,
        ),
    )

    if expect_raises:
        with pytest.raises(expect_raises):
            run_plan()
    else:
        run_plan()

    mock_level_set = get_mock_put(oav_full_screen.zoom_controller.level)
    mock_level_set.assert_has_calls([call("1.0x"), call("2.0x")])


@patch("mx_bluesky.beamlines.i04.thawing_plan.MurkoCallback")
def test_given_thaw_succeeds_then_thaw_and_stream_sets_zoom_to_1_and_back(
    patch_murko_callback: MagicMock,
    smargon: Smargon,
    thawer: Thawer,
    oav_forwarder: OAVToRedisForwarder,
    oav_full_screen: OAV,
    robot: BartRobot,
    run_engine: RunEngine,
):
    _run_thaw_and_stream_and_assert_zoom_changes(
        smargon, thawer, oav_forwarder, oav_full_screen, robot, run_engine
    )


@patch("mx_bluesky.beamlines.i04.thawing_plan.MurkoCallback")
@patch("mx_bluesky.beamlines.i04.thawing_plan.bps.monitor")
def test_given_thaw_fails_then_thaw_and_stream_sets_zoom_to_1_and_back(
    mock__thaw: MagicMock,
    patch_murko_callback: MagicMock,
    smargon: Smargon,
    thawer: Thawer,
    oav_forwarder: OAVToRedisForwarder,
    oav_full_screen: OAV,
    robot: BartRobot,
    run_engine: RunEngine,
):
    mock__thaw.side_effect = Exception()
    _run_thaw_and_stream_and_assert_zoom_changes(
        smargon, thawer, oav_forwarder, oav_full_screen, robot, run_engine, Exception
    )


@patch("mx_bluesky.beamlines.i04.thawing_plan.MurkoCallback")
@patch(
    "mx_bluesky.beamlines.i04.thawing_plan._rotate_in_one_direction_and_stream_to_redis"
)
def test_thaw_and_murko_centre_stages_and_unstages_murko_results_twice(
    mock_rotate_and_stream: MagicMock,
    patch_murko_callback: MagicMock,
    smargon: Smargon,
    thawer: Thawer,
    oav_forwarder: OAVToRedisForwarder,
    robot: BartRobot,
    murko_results: MurkoResultsDevice,
    run_engine: RunEngine,
):
    run_engine(
        thaw_and_murko_centre(
            10, 360, robot, thawer, smargon, murko_results, oav_forwarder
        ),
    )
    assert murko_results.stage.call_count == 2  # type: ignore
    assert murko_results.unstage.call_count == 2  # type: ignore


@patch("mx_bluesky.beamlines.i04.thawing_plan.MurkoCallback")
@patch("mx_bluesky.beamlines.i04.thawing_plan.bps.monitor")
def test_given_thaw_and_murko_centre_errors_then_murko_results_still_unstaged(
    mock__thaw: MagicMock,
    patch_murko_callback: MagicMock,
    smargon: Smargon,
    thawer: Thawer,
    oav_forwarder: OAVToRedisForwarder,
    robot: BartRobot,
    murko_results: MurkoResultsDevice,
    run_engine: RunEngine,
):
    mock__thaw.side_effect = ValueError()

    with pytest.raises(ValueError):
        run_engine(
            thaw_and_murko_centre(
                10, 360, robot, thawer, smargon, murko_results, oav_forwarder
            ),
        )

    murko_results.stage.assert_called_once()  # type: ignore
    murko_results.unstage.assert_called_once()  # type: ignore


def test_thaw_and_murko_centre_will_switch_murko_source_half_way_through_thaw(
    sim_run_engine,
    smargon: Smargon,
    thawer: Thawer,
    oav_forwarder: OAVToRedisForwarder,
    robot: BartRobot,
    murko_results: MurkoResultsDevice,
):
    _test_plan_will_switch_murko_source_half_way_through_thaw(
        sim_run_engine,
        thaw_and_murko_centre(
            10, 360, robot, thawer, smargon, murko_results, oav_forwarder
        ),
    )


@patch("mx_bluesky.beamlines.i04.thawing_plan.MurkoCallback")
def test_thaw_and_murko_centre_will_centre_based_on_murko_results_after_both_rotations(
    patch_murko_callback,
    run_engine,
    smargon: Smargon,
    thawer: Thawer,
    oav_forwarder: OAVToRedisForwarder,
    robot: BartRobot,
    murko_results: MurkoResultsDevice,
):
    def fake_trigger(call_count):
        if call_count == 1:
            murko_results._x_mm_setter(1)
            murko_results._y_mm_setter(2)
            murko_results._z_mm_setter(3)
        elif call_count == 2:
            murko_results._x_mm_setter(4)  # user setpoint = 1 + 4 = 5
            murko_results._y_mm_setter(5)  # user setpoint = 2 + 5 = 7
            murko_results._z_mm_setter(6)  # user setpoint = 3 + 6 = 9
        else:
            raise Exception  # should only trigger twice
        return completed_status()

    mock_trigger = MagicMock()
    murko_results.trigger = mock_trigger

    def side_effect():
        return fake_trigger(mock_trigger.call_count)

    mock_trigger.side_effect = side_effect

    run_engine(
        thaw_and_murko_centre(
            10, 360, robot, thawer, smargon, murko_results, oav_forwarder
        ),
    )

    get_mock_put(smargon.x.user_setpoint).assert_has_calls([call(1.0)])
    get_mock_put(smargon.y.user_setpoint).assert_has_calls([call(2.0)])
    get_mock_put(smargon.z.user_setpoint).assert_has_calls([call(3.0)])
    get_mock_put(smargon.x.user_setpoint).assert_has_calls([call(5.0)])
    get_mock_put(smargon.y.user_setpoint).assert_has_calls([call(7.0)])
    get_mock_put(smargon.z.user_setpoint).assert_has_calls([call(9.0)])


def test_thaw_and_murko_centre_will_set_sample_id_before_triggering_results(
    sim_run_engine,
    smargon: Smargon,
    thawer: Thawer,
    oav_forwarder: OAVToRedisForwarder,
    robot: BartRobot,
    murko_results: MurkoResultsDevice,
):
    sim_run_engine.add_read_handler_for(robot.sample_id, "1234")

    msgs = sim_run_engine.simulate_plan(
        thaw_and_murko_centre(
            10, 360, robot, thawer, smargon, murko_results, oav_forwarder
        )
    )

    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == "murko_results-sample_id"
        and msg.args[0] == "1234",
    )
    msgs = assert_message_and_return_remaining(
        msgs, lambda msg: msg.command == "trigger" and msg.obj.name == "murko_results"
    )


def test_thawing_plan_with_murko_callback_puts_correct_metadata_into_redis(
    run_engine,
    smargon: Smargon,
    thawer: Thawer,
    oav_forwarder: OAVToRedisForwarder,
    robot: BartRobot,
    murko_results: MurkoResultsDevice,
    murko_callback: MurkoCallback,
):
    with patch(
        "mx_bluesky.beamlines.i04.thawing_plan.MurkoCallback"
    ) as mock_murko_callback:
        mock_murko_callback.return_value = murko_callback
        run_engine(
            thaw_and_murko_centre(
                10, 360, robot, thawer, smargon, murko_results, oav_forwarder
            ),
        )

    expected_full_screen_md = {
        "sample_id": 0,
        "microns_per_x_pixel": 2.87,
        "microns_per_y_pixel": 2.8699999999999997,
        "beam_centre_i": 0,
        "beam_centre_j": 0,
        "omega_angle": 360.0,
        "uuid": "test",
    }
    expected_roi_md = {
        "sample_id": 0,
        "microns_per_x_pixel": 3.16,
        "microns_per_y_pixel": 3.16,
        "beam_centre_i": 0,
        "beam_centre_j": 0,
        "omega_angle": 0.0,
        "uuid": "test",
    }

    hset_call_args_list = murko_callback.redis_client.hset.call_args_list  # type: ignore
    assert len(hset_call_args_list) == 2
    assert hset_call_args_list[0].args[2] == json.dumps(expected_full_screen_md)
    assert hset_call_args_list[1].args[2] == json.dumps(expected_roi_md)

    publish_call_args_list = murko_callback.redis_client.publish.call_args_list  # type: ignore
    assert len(publish_call_args_list) == 4
    assert publish_call_args_list[0].args[1] == json.dumps(expected_full_screen_md)
    assert publish_call_args_list[1].args[1] == json.dumps(FORWARDING_COMPLETE_MESSAGE)
    assert publish_call_args_list[2].args[1] == json.dumps(expected_roi_md)
    assert publish_call_args_list[3].args[1] == json.dumps(FORWARDING_COMPLETE_MESSAGE)


@patch("mx_bluesky.beamlines.i04.thawing_plan.MurkoCallback._check_redis_connection")
def test_plans_carry_on_thaw_if_redis_connection_check_fails(
    patch_callback_check_redis_connection: MagicMock,
    smargon: Smargon,
    thawer: Thawer,
    robot: BartRobot,
    oav_forwarder: OAVToRedisForwarder,
    run_engine: RunEngine,
):
    patch_callback_check_redis_connection.return_value = False
    murko_results = MurkoResultsDevice()
    murko_results._check_redis_connection = AsyncMock(return_value=False)
    for plan in (
        thaw_and_murko_centre(
            10,
            360,
            thawer=thawer,
            smargon=smargon,
            robot=robot,
            oav_to_redis_forwarder=oav_forwarder,
            murko_results=murko_results,
        ),
        thaw_and_stream_to_redis(
            10,
            360,
            thawer=thawer,
            smargon=smargon,
            robot=robot,
            oav_to_redis_forwarder=oav_forwarder,
        ),
    ):
        run_engine(plan)

        omega_put = get_mock_put(smargon.omega.user_setpoint)

        assert omega_put.call_args_list == [
            call(360.0),
            call(0.0),
        ]

        omega_put.reset_mock()



from mx_bluesky.beamlines.aithre_lasershaping.beamline_safe import (
    go_to_zero,
    set_beamline_safe_on_robot,
)
from mx_bluesky.beamlines.aithre_lasershaping.check_goniometer_performance import (
    check_omega_performance,
)
from mx_bluesky.beamlines.aithre_lasershaping.goniometer_controls import (
    change_goniometer_turn_speed,
    go_to_furthest_maximum,
    jog_sample,
    rotate_goniometer_relative,
)

__all__ = [
    "set_beamline_safe_on_robot",
    "check_omega_performance",
    "change_goniometer_turn_speed",
    "go_to_furthest_maximum",
    "rotate_goniometer_relative",
    "jog_sample",
    "go_to_zero",
]



from functools import partial

from bluesky import plan_stubs as bps
from bluesky.plans import rel_grid_scan
from bluesky.utils import short_uid
from dodal.beamlines.i23 import I23DetectorPositions
from dodal.common import inject
from dodal.devices.motors import SixAxisGonio
from dodal.devices.positioner import Positioner1D
from ophyd_async.epics.motor import Motor


def set_axis_to_max_velocity(axis: Motor):
    max_vel = yield from bps.rd(axis.max_velocity)
    yield from bps.mv(axis.velocity, max_vel)


def one_nd_step(
    detectors,
    step,
    pos_cache,
    omega_axis: Motor,
    omega_rotation: float,
    omega_velocity: float,
):
    def move():
        yield from bps.checkpoint()
        grp = short_uid("set")
        for motor, pos in step.items():
            yield from bps.abs_set(motor, pos, group=grp)
        yield from set_axis_to_max_velocity(omega_axis)
        yield from bps.abs_set(omega_axis, 0, group=grp)
        yield from bps.wait(group=grp)

    yield from move()
    yield from bps.mv(omega_axis.velocity, omega_velocity)
    yield from bps.mv(omega_axis, omega_rotation)


def serial_collection(
    x_steps: int,
    y_steps: int,
    x_step_size: float,
    y_step_size: float,
    omega_rotation: float,
    omega_velocity: float,
    detector_motion: Positioner1D = inject("detector_motion"),
    gonio: SixAxisGonio = inject("gonio"),
):
    """This plan runs a software controlled serial collection. i.e it moves in a snaked
    grid and does a small rotation collection at each point."""

    yield from bps.mv(detector_motion.stage_position, I23DetectorPositions.IN)
    yield from rel_grid_scan(
        [],
        gonio.y,
        0,
        y_step_size * (y_steps - 1),
        y_steps,
        gonio.x,
        0,
        x_step_size * (x_steps - 1),
        x_steps,
        per_step=partial(  # type: ignore
            one_nd_step,
            omega_axis=gonio.omega,
            omega_rotation=omega_rotation,
            omega_velocity=omega_velocity,
        ),
        snake_axes=True,
    )



#!/usr/bin/env python3
import argparse
from inspect import get_annotations, getmodule, getmro, isclass
from typing import get_type_hints

from blueapi.core.bluesky_types import is_bluesky_plan_generator
from blueapi.core.context import load_module_all
from pydantic import BaseModel

from mx_bluesky.common.parameters.components import MxBlueskyParameters
from mx_bluesky.hyperion import experiment_plans


def main():
    parser = argparse.ArgumentParser(
        description="Generate PlantUML diagram for Hyperion parameter models."
    )
    parser.add_argument(
        "outfile",
        help="Path to the output .puml file",
    )
    args = parser.parse_args()

    with open(args.outfile, "w") as out:
        write_header(out)

        parameter_types = [
            get_type_hints(obj)["parameters"]
            for obj in load_module_all(experiment_plans)
            if is_bluesky_plan_generator(obj)
        ]

        experiment_types = set(parameter_types)
        all_types = set(parameter_types)

        # Collect all BaseModel ancestors
        for experiment_type in experiment_types:
            for base in get_mro_base_models(experiment_type):
                all_types.add(base)

        # Group by package
        types_by_package = group_types_by_package(all_types)

        # Split mx_bluesky.common.parameters into mixins and main parameter types
        mx_bluesky_param_types, mixin_types = split_mx_bluesky_types(types_by_package)

        # Write mx_bluesky.common.parameters
        out.write("package mx_bluesky.common.parameters {\n")
        out.write("together {\n")
        for t in mixin_types:
            generate_class(t, out)
        out.write("}\n")
        out.write("together {\n")
        for t in mx_bluesky_param_types:
            generate_class(t, out)
        out.write("}\n")
        out.write("}\n")

        # Write mx_bluesky.hyperion.parameters
        out.write("package mx_bluesky.hyperion.parameters {\n")
        for t in types_by_package["mx_bluesky.hyperion.parameters"]:
            generate_class(t, out)
        out.write("}\n")

        # Write inheritance arrows
        for t in all_types:
            for base in t.__bases__:
                if base not in (object, BaseModel):
                    out.write(f"{base.__name__} <|-- {t.__name__}\n")

        out.write("@enduml\n")


def write_header(out):
    out.write(
        "'This file is auto-generated by generate_plantuml.py\n"
        "@startuml hyperion_parameter_model\n"
        "title Hyperion Parameter Model\n"
        "set namespaceSeparator none\n\n"
    )


def get_mro_base_models(t):
    for base in getmro(t):
        if issubclass(base, BaseModel) and base is not BaseModel:
            yield base


def group_types_by_package(all_types):
    grouped = {}
    for t in all_types:
        mod = getmodule(t)
        assert mod
        grouped.setdefault(mod.__package__, []).append(t)
    return grouped


def split_mx_bluesky_types(types_by_package):
    mx_bluesky_param_types = set()
    mixin_types = set()

    for t in types_by_package.get("mx_bluesky.common.parameters", []):
        if issubclass(t, MxBlueskyParameters):
            mx_bluesky_param_types.add(t)
        else:
            mixin_types.add(t)

    return mx_bluesky_param_types, mixin_types


def generate_class(t, out):
    out.write(f"class {t.__name__}{{\n")
    for field_name, field_type in get_annotations(t).items():
        out.write(f"\t{generate_type(field_type)} {field_name}\n")
    out.write("}\n")


def generate_type(field_type):
    return (
        field_type.__name__
        if isclass(field_type) and issubclass(field_type, BaseModel)
        else str(field_type)
    )


if __name__ == "__main__":
    main()



from unittest.mock import MagicMock

import bluesky.preprocessors as bpp
import pytest
from bluesky import plan_stubs as bps
from bluesky.run_engine import RunEngine

from mx_bluesky.common.external_interaction.callbacks.common.ispyb_callback_base import (
    BaseISPyBCallback,
)
from mx_bluesky.common.parameters.constants import USE_NUMTRACKER
from mx_bluesky.common.parameters.gridscan import SpecifiedThreeDGridScan


def test_visit_extracted_from_numtracker(
    run_engine: RunEngine, test_fgs_params: SpecifiedThreeDGridScan
):
    test_visit = "test_visit"

    # BlueAPI does this when submitting a task
    run_engine.md.update({"instrument_session": test_visit})

    callback = BaseISPyBCallback()
    callback.activity_gated_stop = MagicMock()
    test_fgs_params.visit = USE_NUMTRACKER
    callback.params = test_fgs_params
    run_engine.subscribe(callback)

    @bpp.run_decorator(
        md={
            "activate_callbacks": ["BaseISPyBCallback"],
        },
    )
    def test_plan():
        yield from bps.null()

    run_engine(test_plan())

    assert callback.params.visit == test_visit


def test_exception_when_instrument_session_doesnt_exist(
    run_engine: RunEngine, test_fgs_params: SpecifiedThreeDGridScan
):
    callback = BaseISPyBCallback()
    callback.activity_gated_stop = MagicMock()
    test_fgs_params.visit = USE_NUMTRACKER
    callback.params = test_fgs_params
    run_engine.subscribe(callback)

    @bpp.run_decorator(
        md={
            "activate_callbacks": ["BaseISPyBCallback"],
        },
    )
    def test_plan():
        yield from bps.null()

    with pytest.raises(ValueError):
        run_engine(test_plan())



from unittest.mock import AsyncMock, MagicMock, call, patch

import bluesky.plan_stubs as bps
import pytest
from bluesky.run_engine import RunEngine
from bluesky.simulators import RunEngineSimulator, assert_message_and_return_remaining
from dodal.devices.attenuator.attenuator import BinaryFilterAttenuator
from dodal.devices.backlight import Backlight
from dodal.devices.beamlines.i04.beam_centre import CentreEllipseMethod
from dodal.devices.beamlines.i04.max_pixel import MaxPixel
from dodal.devices.mx_phase1.beamstop import Beamstop, BeamstopPositions
from dodal.devices.oav.oav_detector import OAV, ZoomControllerWithBeamCentres
from dodal.devices.robot import BartRobot, PinMounted
from dodal.devices.scintillator import InOut, Scintillator
from dodal.devices.xbpm_feedback import XBPMFeedback
from dodal.devices.zebra.zebra_controlled_shutter import (
    ZebraShutter,
    ZebraShutterControl,
    ZebraShutterState,
)
from ophyd_async.core import (
    AsyncStatus,
    completed_status,
    get_mock_put,
    init_devices,
    set_mock_value,
)

from mx_bluesky.beamlines.i04.oav_centering_plans.oav_imaging import (
    OAV_PREPARE_BEAMLINE_FOR_SCINT_WAIT,
    FindBeamCentresComposite,
    _prepare_beamline_for_scintillator_images,
    find_and_set_beam_centre_at_current_zoom_and_transmission,
    find_beam_centres,
    optimise_transmission_with_oav,
    take_and_save_oav_image,
    take_oav_image_with_scintillator_in,
)
from mx_bluesky.common.utils.exceptions import BeamlineStateError
from tests.conftest import fake_generator


async def test_check_exception_raised_if_pin_mounted(
    run_engine: RunEngine,
    robot: BartRobot,
    beamstop_phase1: Beamstop,
    scintillator: Scintillator,
    attenuator: BinaryFilterAttenuator,
    sample_shutter: ZebraShutter,
    oav: OAV,
):
    set_mock_value(robot.gonio_pin_sensor, PinMounted.PIN_MOUNTED)

    with pytest.raises(BeamlineStateError, match="Pin should not be mounted!"):
        run_engine(
            take_oav_image_with_scintillator_in(
                robot=robot,
                beamstop=beamstop_phase1,
                scintillator=scintillator,
                attenuator=attenuator,
                shutter=sample_shutter,
                oav=oav,
            )
        )


def test_prepare_beamline_for_scint_images(
    sim_run_engine: RunEngineSimulator,
    robot: BartRobot,
    beamstop_phase1: Beamstop,
    backlight: Backlight,
    scintillator: Scintillator,
    xbpm_feedback: XBPMFeedback,
    sample_shutter: ZebraShutter,
):
    test_group = "my_group"
    messages = sim_run_engine.simulate_plan(
        _prepare_beamline_for_scintillator_images(
            robot,
            beamstop_phase1,
            backlight,
            scintillator,
            xbpm_feedback,
            sample_shutter,
            test_group,
        )
    )

    messages = assert_message_and_return_remaining(
        messages,
        lambda msg: msg.command == "read" and msg.obj.name == "robot-gonio_pin_sensor",
    )

    messages = assert_message_and_return_remaining(
        messages,
        lambda msg: msg.command == "trigger" and msg.obj.name == "xbpm_feedback",
    )

    messages = assert_message_and_return_remaining(
        messages,
        lambda msg: msg.command == "set"
        and msg.obj.name == "beamstop-selected_pos"
        and msg.args[0] == BeamstopPositions.DATA_COLLECTION
        and msg.kwargs["group"] == test_group,
    )

    messages = assert_message_and_return_remaining(
        messages,
        lambda msg: msg.command == "set"
        and msg.obj.name == "backlight"
        and msg.args[0] == InOut.OUT,
    )

    messages = assert_message_and_return_remaining(
        messages,
        lambda msg: msg.command == "set"
        and msg.obj.name == "scintillator-selected_pos"
        and msg.args[0] == InOut.IN
        and msg.kwargs["group"] == test_group,
    )

    messages = assert_message_and_return_remaining(
        messages,
        lambda msg: msg.command == "set"
        and msg.obj.name == "sample_shutter-control_mode"
        and msg.args[0] == ZebraShutterControl.MANUAL,
    )
    messages = assert_message_and_return_remaining(
        messages, lambda msg: msg.command == "wait"
    )
    assert_message_and_return_remaining(
        messages,
        lambda msg: msg.command == "set"
        and msg.obj.name == "sample_shutter"
        and msg.args[0] == ZebraShutterState.OPEN
        and msg.kwargs["group"] == test_group,
    )


def test_plan_stubs_called_in_correct_order(
    sim_run_engine: RunEngineSimulator,
    robot: BartRobot,
    beamstop_phase1: Beamstop,
    scintillator: Scintillator,
    attenuator: BinaryFilterAttenuator,
    oav: OAV,
    sample_shutter: ZebraShutter,
    backlight: Backlight,
    xbpm_feedback: XBPMFeedback,
):
    messages = sim_run_engine.simulate_plan(
        take_oav_image_with_scintillator_in(
            attenuator=attenuator,
            shutter=sample_shutter,
            oav=oav,
            robot=robot,
            beamstop=beamstop_phase1,
            backlight=backlight,
            scintillator=scintillator,
            xbpm_feedback=xbpm_feedback,
        )
    )

    messages = assert_message_and_return_remaining(
        messages,
        lambda msg: msg.command == "set"
        and msg.obj.name == "attenuator"
        and msg.args[0] == 1,
    )

    messages = assert_message_and_return_remaining(
        messages,
        lambda msg: msg.command == "wait",
    )

    messages = assert_message_and_return_remaining(
        messages,
        lambda msg: msg.command == "set"
        and msg.obj.name == "sample_shutter-control_mode"
        and msg.args[0] == ZebraShutterControl.MANUAL,
    )

    messages = assert_message_and_return_remaining(
        messages,
        lambda msg: msg.command == "wait"
        and msg.kwargs["group"] == messages[0].kwargs["group"],
    )

    messages = assert_message_and_return_remaining(
        messages,
        lambda msg: msg.command == "set"
        and msg.obj.name == "sample_shutter"
        and msg.args[0] == ZebraShutterState.OPEN,
    )

    messages = assert_message_and_return_remaining(
        messages,
        lambda msg: msg.command == "wait"
        and msg.kwargs["group"] == messages[0].kwargs["group"],
    )


@pytest.mark.parametrize(
    "transmission",
    [1, 0.5, 0.1],
)
def test_plan_called_with_specified_transmission_then_transmission_set(
    sim_run_engine: RunEngineSimulator,
    robot: BartRobot,
    beamstop_phase1: Beamstop,
    scintillator: Scintillator,
    attenuator: BinaryFilterAttenuator,
    oav: OAV,
    sample_shutter: ZebraShutter,
    backlight: Backlight,
    xbpm_feedback: XBPMFeedback,
    transmission: float,
):
    messages = sim_run_engine.simulate_plan(
        take_oav_image_with_scintillator_in(
            transmission=transmission,
            attenuator=attenuator,
            shutter=sample_shutter,
            oav=oav,
            robot=robot,
            beamstop=beamstop_phase1,
            backlight=backlight,
            scintillator=scintillator,
            xbpm_feedback=xbpm_feedback,
        )
    )

    messages = assert_message_and_return_remaining(
        messages,
        lambda msg: msg.command == "set"
        and msg.obj.name == "attenuator"
        and msg.args[0] == transmission,
    )


def test_oav_image(sim_run_engine: RunEngineSimulator, oav: OAV):
    mock_filepath = "mock_path"
    mock_filename = "mock_file"
    messages = sim_run_engine.simulate_plan(
        take_and_save_oav_image(mock_filename, mock_filepath, oav)
    )

    messages = assert_message_and_return_remaining(
        messages,
        lambda msg: msg.command == "set"
        and msg.obj.name == "oav-snapshot-filename"
        and msg.args[0] == mock_filename,
    )

    messages = assert_message_and_return_remaining(
        messages,
        lambda msg: msg.command == "set"
        and msg.obj.name == "oav-snapshot-directory"
        and msg.args[0] == mock_filepath,
    )

    messages = assert_message_and_return_remaining(
        messages,
        lambda msg: msg.command == "wait",
    )

    messages = assert_message_and_return_remaining(
        messages,
        lambda msg: msg.command == "trigger" and msg.obj.name == "oav-snapshot",
    )


@patch(
    "mx_bluesky.beamlines.i04.oav_centering_plans.oav_imaging.os.path.exists",
    MagicMock(return_value=True),
)
def test_given_file_exists_then_take_oav_image_raises(
    sim_run_engine: RunEngineSimulator, oav: OAV
):
    with pytest.raises(FileExistsError):
        sim_run_engine.simulate_plan(
            take_and_save_oav_image("mock_file", "mock_path", oav)
        )


async def test_take_and_save_oav_image_in_re(run_engine: RunEngine, oav: OAV, tmp_path):
    expected_filename = "filename"
    expected_directory = str(tmp_path)
    run_engine(take_and_save_oav_image(expected_filename, expected_directory, oav))
    assert await oav.snapshot.filename.get_value() == expected_filename
    assert await oav.snapshot.directory.get_value() == str(expected_directory)
    oav.snapshot.trigger.assert_called_once()  # type: ignore


@pytest.fixture()
async def max_pixel() -> MaxPixel:
    async with init_devices(mock=True):
        max_pixel = MaxPixel("TEST: MAX_PIXEL", "max_pixel")

    max_pixel.trigger = MagicMock(return_value=completed_status())
    return max_pixel


def test_optimise_transmission_first_gets_max_pixel_at_100_percent(
    sim_run_engine: RunEngineSimulator,
    attenuator: BinaryFilterAttenuator,
    xbpm_feedback: XBPMFeedback,
    max_pixel: MaxPixel,
):
    max_values = [100, 75]

    def return_max_values(_):
        return {"readback": {"value": max_values.pop(0)}}

    sim_run_engine.add_handler("read", return_max_values, max_pixel.max_pixel_val.name)

    messages = sim_run_engine.simulate_plan(
        optimise_transmission_with_oav(
            max_pixel=max_pixel,
            attenuator=attenuator,
            xbpm_feedback=xbpm_feedback,
        )
    )

    messages = assert_message_and_return_remaining(
        messages,
        lambda msg: msg.command == "trigger" and msg.obj == xbpm_feedback,
    )

    messages = assert_message_and_return_remaining(
        messages,
        lambda msg: msg.command == "set" and msg.obj == attenuator and msg.args[0] == 1,
    )

    messages = assert_message_and_return_remaining(
        messages, lambda msg: msg.command == "trigger" and msg.obj == max_pixel
    )

    messages = assert_message_and_return_remaining(
        messages,
        lambda msg: msg.command == "read" and msg.obj == max_pixel.max_pixel_val,
    )


@pytest.mark.parametrize("iterations", [10, 6, 4])
def test_given_transmission_change_always_high_then_raises_stop_iteration(
    attenuator: BinaryFilterAttenuator,
    xbpm_feedback: XBPMFeedback,
    max_pixel: MaxPixel,
    run_engine: RunEngine,
    iterations: int,
):
    set_mock_value(max_pixel.max_pixel_val, 100)

    @AsyncStatus.wrap
    async def fake_attenuator_set(val):
        set_mock_value(attenuator.actual_transmission, val + 0.2)

    attenuator.set = MagicMock(side_effect=fake_attenuator_set)

    with pytest.raises(RuntimeError) as e:
        run_engine(
            optimise_transmission_with_oav(
                max_pixel=max_pixel,
                attenuator=attenuator,
                xbpm_feedback=xbpm_feedback,
                max_iterations=iterations,
            )
        )

    # The RE hides the StopIteration behind a RuntimeError but will mention it in the message
    assert "StopIteration" in e.value.args[0]
    assert max_pixel.trigger.call_count == iterations + 1  # type: ignore


def given_max_values(max_pixel: MaxPixel, max_values: list):
    def _set_max_value():
        set_mock_value(max_pixel.max_pixel_val, max_values.pop(0))
        return completed_status()

    max_pixel.trigger.side_effect = _set_max_value  # type: ignore


@pytest.mark.parametrize(
    "lower_bound, upper_bound, expected_final_transmission",
    [[0, 100, 50], [0, 10, 5], [5, 25, 15]],
)
async def test_given_max_pixel_immediately_reaches_target_then_optimise_transmission_returns_half_bounds(
    attenuator: BinaryFilterAttenuator,
    xbpm_feedback: XBPMFeedback,
    max_pixel: MaxPixel,
    run_engine: RunEngine,
    lower_bound: int,
    upper_bound: int,
    expected_final_transmission: int,
):
    given_max_values(max_pixel, [100, 75])

    run_engine(
        optimise_transmission_with_oav(
            lower_bound=lower_bound,
            upper_bound=upper_bound,
            max_pixel=max_pixel,
            attenuator=attenuator,
            xbpm_feedback=xbpm_feedback,
        )
    )

    final_transmission = (await attenuator.actual_transmission.get_value()) * 100
    assert final_transmission == expected_final_transmission

    assert attenuator.set.call_args_list == [  # type: ignore
        call(1),
        call(expected_final_transmission / 100),
    ]


@pytest.mark.parametrize(
    "target_fraction",
    [0.75, 0.26, 0.39],
)
async def test_optimise_transmission_reaches_different_target_fractions(
    attenuator: BinaryFilterAttenuator,
    xbpm_feedback: XBPMFeedback,
    max_pixel: MaxPixel,
    run_engine: RunEngine,
    target_fraction: float,
):
    given_max_values(max_pixel, [100, 100 * target_fraction])

    run_engine(
        optimise_transmission_with_oav(
            target_brightness_fraction=target_fraction,
            max_pixel=max_pixel,
            attenuator=attenuator,
            xbpm_feedback=xbpm_feedback,
        )
    )

    final_transmission = (await attenuator.actual_transmission.get_value()) * 100
    assert final_transmission == 50

    assert attenuator.set.call_args_list == [call(1), call(0.5)]  # type: ignore


async def test_max_pixel_stays_too_large_then_optimise_transmission_keeps_reducing(
    attenuator: BinaryFilterAttenuator,
    xbpm_feedback: XBPMFeedback,
    max_pixel: MaxPixel,
    run_engine: RunEngine,
):
    given_max_values(max_pixel, [100, 100, 100, 100, 100, 75])

    run_engine(
        optimise_transmission_with_oav(
            max_pixel=max_pixel,
            attenuator=attenuator,
            xbpm_feedback=xbpm_feedback,
        )
    )

    final_transmission = (await attenuator.actual_transmission.get_value()) * 100
    assert final_transmission == 6.25

    assert attenuator.set.call_args_list == [  # type: ignore
        call(1),
        call(0.5),
        call(pytest.approx(0.25)),
        call(pytest.approx(0.125)),
        call(pytest.approx(0.0625)),
    ]


async def test_max_pixel_stays_too_small_then_optimise_transmission_keeps_increasing(
    attenuator: BinaryFilterAttenuator,
    xbpm_feedback: XBPMFeedback,
    max_pixel: MaxPixel,
    run_engine: RunEngine,
):
    given_max_values(max_pixel, [100, 20, 20, 20, 20, 75])

    run_engine(
        optimise_transmission_with_oav(
            max_pixel=max_pixel,
            attenuator=attenuator,
            xbpm_feedback=xbpm_feedback,
        )
    )

    final_transmission = (await attenuator.actual_transmission.get_value()) * 100
    assert final_transmission == 93.75

    assert attenuator.set.call_args_list == [  # type: ignore
        call(1),
        call(0.5),
        call(pytest.approx(0.75)),
        call(pytest.approx(0.875)),
        call(pytest.approx(0.9375)),
    ]


@pytest.mark.parametrize(
    "min_trans_change, expected_final_transmission, expected_calls",
    [
        (30, 50.0, [call(1), call(0.5)]),
        (15, 75.0, [call(1), call(0.5), call(0.75)]),
    ],
)
async def test_different_min_trans_change_change_when_we_accept(
    attenuator: BinaryFilterAttenuator,
    xbpm_feedback: XBPMFeedback,
    max_pixel: MaxPixel,
    run_engine: RunEngine,
    min_trans_change: int,
    expected_final_transmission: float,
    expected_calls: list,
):
    given_max_values(max_pixel, [100, 68, 75])

    run_engine(
        optimise_transmission_with_oav(
            min_transmission_change=min_trans_change,
            max_pixel=max_pixel,
            attenuator=attenuator,
            xbpm_feedback=xbpm_feedback,
        )
    )

    final_transmission = (await attenuator.actual_transmission.get_value()) * 100

    assert final_transmission == expected_final_transmission
    assert attenuator.set.call_args_list == expected_calls  # type: ignore


async def test_brightness_alternates_above_then_below_target_bounds_shrink_both_sides(
    attenuator: BinaryFilterAttenuator,
    xbpm_feedback: XBPMFeedback,
    max_pixel: MaxPixel,
    run_engine: RunEngine,
):
    given_max_values(max_pixel, [100, 90, 60, 85, 65, 75])

    run_engine(
        optimise_transmission_with_oav(
            max_pixel=max_pixel,
            attenuator=attenuator,
            xbpm_feedback=xbpm_feedback,
        )
    )

    final_transmission = (await attenuator.actual_transmission.get_value()) * 100
    assert final_transmission == 31.25

    # Note the 2 dp rounding on set values:
    assert attenuator.set.call_args_list == [  # type: ignore
        call(1),
        call(0.5),
        call(0.25),
        call(pytest.approx(0.375)),
        call(pytest.approx(0.3125)),
    ]


async def test_equal_to_target_matches_target(
    attenuator: BinaryFilterAttenuator,
    xbpm_feedback: XBPMFeedback,
    max_pixel: MaxPixel,
    run_engine: RunEngine,
):
    given_max_values(max_pixel, [100, 75])

    run_engine(
        optimise_transmission_with_oav(
            max_pixel=max_pixel,
            attenuator=attenuator,
            xbpm_feedback=xbpm_feedback,
            target_brightness_fraction=0.75,
        )
    )

    final_transmission = (await attenuator.actual_transmission.get_value()) * 100
    assert final_transmission == 50
    assert attenuator.set.call_args_list == [call(1), call(0.5)]  # type:ignore


def test_optimise_transmission_raises_value_error_when_upper_bound_less_than_lower_bound(
    attenuator: BinaryFilterAttenuator,
    xbpm_feedback: XBPMFeedback,
    max_pixel: MaxPixel,
    run_engine: RunEngine,
):
    with pytest.raises(ValueError) as excinfo:
        run_engine(
            optimise_transmission_with_oav(
                lower_bound=60,
                upper_bound=40,
                max_pixel=max_pixel,
                attenuator=attenuator,
                xbpm_feedback=xbpm_feedback,
            )
        )
    assert "Upper bound (40) must be higher than lower bound 60" in str(excinfo.value)

    # Ensure nothing was moved/triggered since the
    assert attenuator.set.call_count == 0  # type: ignore
    assert xbpm_feedback.trigger.call_count == 0  # type: ignore


def test_optimise_transmission_raises_value_error_when_full_beam_brightness_is_zero(
    attenuator: BinaryFilterAttenuator,
    xbpm_feedback: XBPMFeedback,
    max_pixel: MaxPixel,
    run_engine: RunEngine,
):
    given_max_values(max_pixel, [0])

    with pytest.raises(ValueError) as excinfo:
        run_engine(
            optimise_transmission_with_oav(
                max_pixel=max_pixel,
                attenuator=attenuator,
                xbpm_feedback=xbpm_feedback,
            )
        )

    assert "No beam" in str(excinfo.value)

    assert attenuator.set.call_count == 1  # type:ignore


@pytest.fixture()
async def centre_ellipse() -> CentreEllipseMethod:
    async with init_devices(mock=True):
        centre_ellipse = CentreEllipseMethod("", name="centre_ellipse")

    centre_ellipse.trigger = MagicMock(return_value=completed_status())
    return centre_ellipse


def initialise_zoom_centres(
    zoom_controller: ZoomControllerWithBeamCentres, init_values: dict
):
    for i, (level, beam_centre) in enumerate(init_values.items()):
        centre_device = zoom_controller.beam_centres[i]
        set_mock_value(centre_device.level_name, level)
        set_mock_value(centre_device.x_centre, beam_centre[0])
        set_mock_value(centre_device.y_centre, beam_centre[1])


@pytest.fixture()
async def zoom_controller_with_centres() -> ZoomControllerWithBeamCentres:
    async with init_devices(mock=True):
        zoom_controller_with_centres = ZoomControllerWithBeamCentres(
            "", "zoom_controller_with_centres"
        )

    level_names = ["1.0x", "2.0x", "3.0x", "7.5x"]
    initialise_zoom_centres(
        zoom_controller_with_centres, dict.fromkeys(level_names, (0, 0))
    )

    return zoom_controller_with_centres


@pytest.fixture()
def find_beam_centre_devices(
    robot: BartRobot,
    beamstop_phase1: Beamstop,
    backlight: Backlight,
    scintillator: Scintillator,
    xbpm_feedback: XBPMFeedback,
    max_pixel: MaxPixel,
    centre_ellipse: CentreEllipseMethod,
    attenuator: BinaryFilterAttenuator,
    zoom_controller_with_centres: ZoomControllerWithBeamCentres,
    sample_shutter: ZebraShutter,
):
    return FindBeamCentresComposite(
        robot=robot,
        beamstop=beamstop_phase1,
        backlight=backlight,
        scintillator=scintillator,
        xbpm_feedback=xbpm_feedback,
        max_pixel=max_pixel,
        centre_ellipse=centre_ellipse,
        attenuator=attenuator,
        zoom_controller=zoom_controller_with_centres,
        shutter=sample_shutter,
    )


def test_given_levels_to_centre_that_dont_exist_when_find_beam_centres_exception_raised(
    find_beam_centre_devices: FindBeamCentresComposite,
    run_engine: RunEngine,
):
    with pytest.raises(ValueError):
        run_engine(
            find_beam_centres(
                zoom_levels_to_centre=("bad_zoom",), composite=find_beam_centre_devices
            )
        )


def test_given_levels_to_optimise_that_dont_exist_when_find_beam_centres_exception_raised(
    find_beam_centre_devices: FindBeamCentresComposite,
    run_engine: RunEngine,
):
    with pytest.raises(ValueError):
        run_engine(
            find_beam_centres(
                zoom_levels_to_optimise_transmission=("bad_zoom",),
                composite=find_beam_centre_devices,
            )
        )


@patch(
    "mx_bluesky.beamlines.i04.oav_centering_plans.oav_imaging._prepare_beamline_for_scintillator_images"
)
@patch(
    "mx_bluesky.beamlines.i04.oav_centering_plans.oav_imaging.optimise_transmission_with_oav",
    new=MagicMock(),
)
def test_find_beam_centres_starts_by_prepping_scintillator(
    mock_prepare_scintillator: AsyncMock,
    find_beam_centre_devices: FindBeamCentresComposite,
    run_engine: RunEngine,
):
    run_engine(find_beam_centres(composite=find_beam_centre_devices))
    mock_prepare_scintillator.assert_called_once()
    assert find_beam_centre_devices.centre_ellipse.trigger.call_count == 4  # type: ignore


def mock_centre_ellipse_with_given_centres(
    centre_ellipse: CentreEllipseMethod, given_centres: list[tuple[int, int]]
):
    def centre_ellipse_trigger_side_effect(*args):
        next_centre = given_centres.pop(0)
        set_mock_value(centre_ellipse.center_x_val, next_centre[0])
        set_mock_value(centre_ellipse.center_y_val, next_centre[1])
        return completed_status()

    centre_ellipse.trigger.side_effect = centre_ellipse_trigger_side_effect  # type:ignore


@patch(
    "mx_bluesky.beamlines.i04.oav_centering_plans.oav_imaging._prepare_beamline_for_scintillator_images",
    new=MagicMock(),
)
@patch(
    "mx_bluesky.beamlines.i04.oav_centering_plans.oav_imaging.optimise_transmission_with_oav",
    new=MagicMock(),
)
async def test_find_beam_centres_iterates_and_sets_centres(
    find_beam_centre_devices: FindBeamCentresComposite,
    run_engine: RunEngine,
):
    level_names = ["1.0x", "2.0x", "3.0x", "7.5x"]
    new_centres = [(100, 100), (200, 200), (300, 300), (400, 400)]
    expected_centres = new_centres.copy()

    centre_ellipse = find_beam_centre_devices.centre_ellipse
    zoom_controller_with_centres: ZoomControllerWithBeamCentres = (
        find_beam_centre_devices.zoom_controller
    )

    mock_centre_ellipse_with_given_centres(centre_ellipse, new_centres)

    run_engine(find_beam_centres(composite=find_beam_centre_devices))

    assert get_mock_put(zoom_controller_with_centres.level).call_count == 4
    assert get_mock_put(zoom_controller_with_centres.level).call_args_list == [
        call(level) for level in level_names
    ]

    for i, centre in zoom_controller_with_centres.beam_centres.items():
        level_name = await centre.level_name.get_value()
        if level_name:
            assert level_name == level_names[i]
            assert (await centre.x_centre.get_value()) == expected_centres[i][0]
            assert (await centre.y_centre.get_value()) == expected_centres[i][1]


@patch(
    "mx_bluesky.beamlines.i04.oav_centering_plans.oav_imaging._prepare_beamline_for_scintillator_images",
    new=MagicMock(),
)
@patch(
    "mx_bluesky.beamlines.i04.oav_centering_plans.oav_imaging.optimise_transmission_with_oav",
    new=MagicMock(),
)
async def test_if_only_some_levels_given_then_find_beam_centres_iterates_and_sets_those_centres(
    find_beam_centre_devices: FindBeamCentresComposite,
    run_engine: RunEngine,
):
    new_centres = [(100, 100), (200, 200), (300, 300), (400, 400)]

    centre_ellipse = find_beam_centre_devices.centre_ellipse
    zoom_controller_with_centres: ZoomControllerWithBeamCentres = (
        find_beam_centre_devices.zoom_controller
    )

    mock_centre_ellipse_with_given_centres(centre_ellipse, new_centres)

    run_engine(
        find_beam_centres(
            zoom_levels_to_centre=("1.0x", "7.5x"), composite=find_beam_centre_devices
        )
    )

    assert get_mock_put(zoom_controller_with_centres.level).call_count == 2
    assert get_mock_put(zoom_controller_with_centres.level).call_args_list == [
        call(level) for level in ("1.0x", "7.5x")
    ]

    centres = list(zoom_controller_with_centres.beam_centres.values())

    assert (await centres[0].level_name.get_value()) == "1.0x"
    assert (await centres[0].x_centre.get_value()) == 100
    assert (await centres[0].y_centre.get_value()) == 100

    for centre in [centres[1], centres[2]]:
        assert (await centre.x_centre.get_value()) == 0
        assert (await centre.y_centre.get_value()) == 0

    assert (await centres[3].level_name.get_value()) == "7.5x"
    assert (await centres[3].x_centre.get_value()) == 200
    assert (await centres[3].y_centre.get_value()) == 200


@patch(
    "mx_bluesky.beamlines.i04.oav_centering_plans.oav_imaging._prepare_beamline_for_scintillator_images",
    new=MagicMock(),
)
@patch(
    "mx_bluesky.beamlines.i04.oav_centering_plans.oav_imaging.optimise_transmission_with_oav"
)
async def test_find_beam_centres_optimises_on_default_levels_only(
    mock_optimise: MagicMock,
    find_beam_centre_devices: FindBeamCentresComposite,
    run_engine: RunEngine,
):
    levels_where_optimised = []

    zoom_controller_with_centres: ZoomControllerWithBeamCentres = (
        find_beam_centre_devices.zoom_controller
    )

    def append_current_zoom(*args, **kwargs):
        current_zoom = yield from bps.rd(zoom_controller_with_centres.level)
        levels_where_optimised.append(current_zoom)
        yield from bps.null()

    mock_optimise.side_effect = append_current_zoom

    run_engine(find_beam_centres(composite=find_beam_centre_devices))

    assert mock_optimise.call_count == 2
    assert levels_where_optimised == ["1.0x", "7.5x"]


@patch(
    "mx_bluesky.beamlines.i04.oav_centering_plans.oav_imaging._prepare_beamline_for_scintillator_images",
    new=MagicMock(),
)
@patch(
    "mx_bluesky.beamlines.i04.oav_centering_plans.oav_imaging.optimise_transmission_with_oav"
)
async def test_find_beam_centres_respects_custom_optimise_list(
    mock_optimise: MagicMock,
    find_beam_centre_devices: FindBeamCentresComposite,
    run_engine: RunEngine,
):
    levels_where_optimised = []

    zoom_controller_with_centres: ZoomControllerWithBeamCentres = (
        find_beam_centre_devices.zoom_controller
    )

    def append_current_zoom(*args, **kwargs):
        current_zoom = yield from bps.rd(zoom_controller_with_centres.level)
        levels_where_optimised.append(current_zoom)
        yield from bps.null()

    mock_optimise.side_effect = append_current_zoom

    run_engine(
        find_beam_centres(
            zoom_levels_to_optimise_transmission=("2.0x", "3.0x"),
            composite=find_beam_centre_devices,
        )
    )

    assert mock_optimise.call_count == 2
    assert levels_where_optimised == ["2.0x", "3.0x"]


@patch(
    "mx_bluesky.beamlines.i04.oav_centering_plans.oav_imaging._get_all_zoom_levels",
    lambda _: fake_generator(("1.0x", "7.5x")),
)
def test_find_beam_centres_prepares_beamline_and_waits(
    sim_run_engine: RunEngineSimulator,
    find_beam_centre_devices: FindBeamCentresComposite,
):
    messages = sim_run_engine.simulate_plan(
        find_beam_centres(composite=find_beam_centre_devices)
    )
    assert_message_and_return_remaining(
        messages,
        lambda msg: msg.command == "set"
        and msg.obj.name == "sample_shutter"
        and msg.args[0] == ZebraShutterState.OPEN
        and msg.kwargs["group"] == OAV_PREPARE_BEAMLINE_FOR_SCINT_WAIT,
    )
    assert_message_and_return_remaining(
        messages,
        lambda msg: msg.command == "wait"
        and msg.kwargs["group"] == OAV_PREPARE_BEAMLINE_FOR_SCINT_WAIT,
    )


@patch(
    "mx_bluesky.beamlines.i04.oav_centering_plans.oav_imaging._prepare_beamline_for_scintillator_images",
    new=MagicMock(),
)
@patch(
    "mx_bluesky.beamlines.i04.oav_centering_plans.oav_imaging.find_beam_centres",
)
def test_find_beam_centre_at_current_zoom_and_transmission_calls_find_beam_centres(
    mock_find_beam_centres: MagicMock,
    find_beam_centre_devices: FindBeamCentresComposite,
    run_engine: RunEngine,
):
    zoom_controller: ZoomControllerWithBeamCentres = (
        find_beam_centre_devices.zoom_controller
    )

    set_mock_value(zoom_controller.level, "2.0x")
    run_engine(
        find_and_set_beam_centre_at_current_zoom_and_transmission(
            find_beam_centre_devices
        )
    )

    mock_find_beam_centres.assert_called_once_with(
        zoom_levels_to_centre=("2.0x",),
        zoom_levels_to_optimise_transmission=(),
        composite=find_beam_centre_devices,
    )


@patch(
    "mx_bluesky.beamlines.i04.oav_centering_plans.oav_imaging._prepare_beamline_for_scintillator_images",
    new=MagicMock(),
)
async def test_find_beam_centre_at_current_zoom_and_transmission_only_finds_centre_for_current_zoom_level(
    find_beam_centre_devices: FindBeamCentresComposite, run_engine: RunEngine
):
    current_zoom_level = "3.0x"
    zoom_controller: ZoomControllerWithBeamCentres = (
        find_beam_centre_devices.zoom_controller
    )
    set_mock_value(zoom_controller.level, current_zoom_level)
    run_engine(
        find_and_set_beam_centre_at_current_zoom_and_transmission(
            find_beam_centre_devices
        )
    )

    get_mock_put(zoom_controller.level).assert_called_once_with(current_zoom_level)
    total_mv_calls = 0
    for centring_device in zoom_controller.beam_centres.values():
        zoom_level = await centring_device.level_name.get_value()
        expected_mv_calls = 1 if zoom_level == current_zoom_level else 0
        assert get_mock_put(centring_device.x_centre).call_count == expected_mv_calls
        assert get_mock_put(centring_device.y_centre).call_count == expected_mv_calls
        total_mv_calls += expected_mv_calls
    assert total_mv_calls == 1


@patch(
    "mx_bluesky.beamlines.i04.oav_centering_plans.oav_imaging._prepare_beamline_for_scintillator_images",
    new=MagicMock(),
)
@patch(
    "mx_bluesky.beamlines.i04.oav_centering_plans.oav_imaging.optimise_transmission_with_oav"
)
def test_find_beam_centre_at_current_zoom_and_transmission_does_not_optimise(
    mock_optimise: MagicMock,
    find_beam_centre_devices: FindBeamCentresComposite,
    run_engine: RunEngine,
):
    zoom_controller: ZoomControllerWithBeamCentres = (
        find_beam_centre_devices.zoom_controller
    )
    set_mock_value(zoom_controller.level, "1.0x")

    run_engine(
        find_and_set_beam_centre_at_current_zoom_and_transmission(
            find_beam_centre_devices
        )
    )

    assert mock_optimise.call_count == 0



from __future__ import annotations

import dataclasses
from collections.abc import Callable, Sequence
from functools import partial

import bluesky.plan_stubs as bps
import bluesky.preprocessors as bpp
from bluesky.protocols import Readable
from bluesky.utils import FailedStatus, MsgGenerator
from dodal.devices.eiger import EigerDetector
from dodal.devices.fast_grid_scan import (
    FastGridScanCommon,
    FastGridScanThreeD,
    GridScanInvalidError,
)

from mx_bluesky.common.experiment_plans.inner_plans.do_fgs import (
    kickoff_and_complete_gridscan,
)
from mx_bluesky.common.experiment_plans.inner_plans.read_hardware import (
    read_hardware_plan,
)
from mx_bluesky.common.parameters.constants import (
    DocDescriptorNames,
    PlanGroupCheckpointConstants,
    PlanNameConstants,
)
from mx_bluesky.common.parameters.device_composites import (
    FlyScanEssentialDevices,
    GonioWithOmegaType,
)
from mx_bluesky.common.parameters.gridscan import SpecifiedThreeDGridScan
from mx_bluesky.common.utils.exceptions import (
    SampleError,
)
from mx_bluesky.common.utils.log import LOGGER
from mx_bluesky.common.utils.tracing import TRACER


@dataclasses.dataclass
class BeamlineSpecificFGSFeatures:
    setup_trigger_plan: Callable[..., MsgGenerator]
    tidy_plan: Callable[..., MsgGenerator]
    set_flyscan_params_plan: Callable[..., MsgGenerator]
    fgs_motors: FastGridScanCommon
    read_pre_flyscan_plan: Callable[
        ..., MsgGenerator
    ]  # Eventually replace with https://github.com/DiamondLightSource/mx-bluesky/issues/819
    read_during_collection_plan: Callable[..., MsgGenerator]


def tidy_eiger(eiger: EigerDetector) -> MsgGenerator:
    """Turn off Eiger dev/shm. Ran after the beamline-specific tidy plan"""

    # Turn off dev/shm streaming to avoid filling disk, see https://github.com/DiamondLightSource/hyperion/issues/1395
    LOGGER.info("Turning off Eiger dev/shm streaming")
    # Fix types in ophyd-async (https://github.com/DiamondLightSource/mx-bluesky/issues/855)
    yield from bps.abs_set(
        eiger.odin.fan.dev_shm_enable,  # type: ignore # until https://github.com/DiamondLightSource/mx-bluesky/issues/1076
        0,
        wait=True,
    )


def construct_beamline_specific_fast_gridscan_features(
    setup_trigger_plan: Callable[..., MsgGenerator],
    tidy_plan: Callable[..., MsgGenerator],
    set_flyscan_params_plan: Callable[..., MsgGenerator],
    fgs_motors: FastGridScanCommon,
    signals_to_read_pre_flyscan: Sequence[Readable],
    signals_to_read_during_collection: Sequence[Readable],
) -> BeamlineSpecificFGSFeatures:
    """Construct the class needed to do beamline-specific parts of the XRC FGS

    Args:
        setup_trigger_plan (Callable): Configure triggering, for example with the Zebra or PandA device.
        Ran directly before kicking off the gridscan.

        tidy_plan (Callable): Tidy up states of devices. Ran at the end of the flyscan, regardless of
        whether or not it finished successfully. Zocalo and Eiger are cleaned up separately

        set_flyscan_params_plan (Callable): Set PV's for the relevant Fast Grid Scan dodal device

        fgs_motors (Callable): Composite device representing the fast grid scan's motion program parameters.

        signals_to_read_pre_flyscan (Callable): Signals which will be read and saved as a bluesky event document
        after all configuration, but before the gridscan.

        signals_to_read_during_collection (Callable): Signals which will be read and saved as a bluesky event
        document whilst the gridscan motion is in progress
    """
    read_pre_flyscan_plan = partial(
        read_hardware_plan,
        signals_to_read_pre_flyscan,
        DocDescriptorNames.HARDWARE_READ_PRE,
    )

    read_during_collection_plan = partial(
        read_hardware_plan,
        signals_to_read_during_collection,
        DocDescriptorNames.HARDWARE_READ_DURING,
    )

    return BeamlineSpecificFGSFeatures(
        setup_trigger_plan,
        tidy_plan,
        set_flyscan_params_plan,
        fgs_motors,
        read_pre_flyscan_plan,
        read_during_collection_plan,
    )


def common_flyscan_xray_centre(
    composite: FlyScanEssentialDevices[GonioWithOmegaType],
    parameters: SpecifiedThreeDGridScan,
    beamline_specific: BeamlineSpecificFGSFeatures,
) -> MsgGenerator:
    """Main entry point of the MX-Bluesky x-ray centering flyscan

    Args:
        composite (FlyScanEssentialDevices): Devices required to perform this plan.

        parameters (SpecifiedThreeDGridScan): Parameters required to perform this plan.

        beamline_specific (BeamlineSpecificFGSFeatures): Configure the beamline-specific version
        of this plan: For example triggering setup and tidy up plans, as well as what to do with the
        centering results.

    With a minimum set of devices and parameters, prepares for; performs; and tidies up a flyscan
    x-ray-center plan. This includes: Configuring desired triggering; writing nexus files; triggering zocalo;
    reading hardware before and during the scan; and tidying up devices after
    the plan is complete. Optionally fetch results from zocalo after completing the grid scan.

    This plan will also push data to ispyb when used with the ispyb_activation_decorator.

    There are a few other useful decorators to use with this plan, see: verify_undulator_gap_before_run_decorator, common/preprocessors/preprocessors.py
    """

    def _overall_tidy():
        yield from beamline_specific.tidy_plan()
        yield from tidy_eiger(composite.eiger)

    def _decorated_flyscan():
        @bpp.set_run_key_decorator(PlanNameConstants.GRIDSCAN_OUTER)
        @bpp.run_decorator(  # attach experiment metadata to the start document
            md={
                "subplan_name": PlanNameConstants.GRIDSCAN_OUTER,
                "mx_bluesky_parameters": parameters.model_dump_json(),
                "activate_callbacks": [
                    "GridscanNexusFileCallback",
                ],
            }
        )
        @bpp.finalize_decorator(lambda: _overall_tidy())
        def run_gridscan_and_tidy(
            fgs_composite: FlyScanEssentialDevices[GonioWithOmegaType],
            params: SpecifiedThreeDGridScan,
            beamline_specific: BeamlineSpecificFGSFeatures,
        ) -> MsgGenerator:
            yield from beamline_specific.setup_trigger_plan(fgs_composite, parameters)

            LOGGER.info("Starting grid scan")
            yield from run_gridscan(fgs_composite, params, beamline_specific)

            LOGGER.info("Grid scan finished")

        yield from run_gridscan_and_tidy(composite, parameters, beamline_specific)

    composite.eiger.set_detector_parameters(parameters.detector_params)
    yield from _decorated_flyscan()


def run_gridscan(
    fgs_composite: FlyScanEssentialDevices[GonioWithOmegaType],
    parameters: SpecifiedThreeDGridScan,
    beamline_specific: BeamlineSpecificFGSFeatures,
):
    # Currently gridscan only works for omega 0, see https://github.com/DiamondLightSource/mx-bluesky/issues/410
    with TRACER.start_span("moving_omega_to_0"):
        yield from bps.abs_set(fgs_composite.gonio.omega, 0)

    with TRACER.start_span("ispyb_hardware_readings"):
        yield from beamline_specific.read_pre_flyscan_plan()

    LOGGER.info("Setting fgs params")

    try:
        yield from beamline_specific.set_flyscan_params_plan()
    except FailedStatus as e:
        if isinstance(e.__cause__, GridScanInvalidError):
            raise SampleError(
                "Scan invalid - gridscan not valid for detected pin position"
            ) from e
        else:
            raise e

    LOGGER.info("Waiting for arming to finish")
    yield from bps.wait(PlanGroupCheckpointConstants.GRID_READY_FOR_DC)
    yield from bps.stage(fgs_composite.eiger, wait=True)

    yield from kickoff_and_complete_gridscan(
        beamline_specific.fgs_motors,
        fgs_composite.eiger,
        fgs_composite.synchrotron,
        [parameters.scan_points_first_grid, parameters.scan_points_second_grid],
        plan_during_collection=beamline_specific.read_during_collection_plan,
    )

    # GDA's 3D gridscans requires Z steps to be at 0, so make sure we leave this device
    # in a GDA-happy state.
    if isinstance(beamline_specific.fgs_motors, FastGridScanThreeD):
        yield from bps.abs_set(beamline_specific.fgs_motors.z_steps, 0, wait=False)



import asyncio
from unittest.mock import MagicMock, patch

import pytest
from bluesky.run_engine import RunEngine
from bluesky.simulators import RunEngineSimulator, assert_message_and_return_remaining
from dodal.devices.beamlines.i24.aperture import AperturePositions
from dodal.devices.beamlines.i24.beamstop import BeamstopPositions
from dodal.devices.beamlines.i24.dual_backlight import BacklightPositions
from dodal.devices.hutch_shutter import ShutterState
from ophyd_async.core import completed_status, set_mock_value

from mx_bluesky.beamlines.i24.jungfrau_commissioning.experiment_plans.rotation_scan_plan import (
    DEFAULT_DETECTOR_DISTANCE_MM,
    JF_DET_STAGE_Y_POSITION_MM,
    ExternalRotationScanParams,
    HutchClosedError,
    RotationScanComposite,
    _cleanup_plan,
    rotation_scan_plan,
    set_up_beamline_for_rotation,
    single_rotation_plan,
)
from mx_bluesky.beamlines.i24.jungfrau_commissioning.plan_stubs.plan_utils import (
    JF_COMPLETE_GROUP,
)
from mx_bluesky.beamlines.i24.parameters.constants import PlanNameConstants
from mx_bluesky.common.experiment_plans.rotation.rotation_utils import (
    calculate_motion_profile,
)
from mx_bluesky.common.parameters.constants import PlanGroupCheckpointConstants
from tests.conftest import raw_params_from_file
from tests.unit_tests.beamlines.i24.jungfrau_commissioning.utils import (
    get_good_single_rotation_params,
)


def get_good_multi_rotation_params(transmissions: list[float], tmp_path):
    params = raw_params_from_file(
        "tests/unit_tests/beamlines/i24/jungfrau_commissioning/test_data/test_good_rotation_params.json",
        tmp_path,
    )
    del params["transmission_frac"]
    params["transmission_fractions"] = transmissions
    params["num_images"] = 1
    return ExternalRotationScanParams(**params)


@patch(
    "mx_bluesky.beamlines.i24.jungfrau_commissioning.experiment_plans.rotation_scan_plan._cleanup_plan"
)
@patch(
    "mx_bluesky.beamlines.i24.jungfrau_commissioning.experiment_plans.rotation_scan_plan.setup_zebra_for_rotation"
)
@patch(
    "mx_bluesky.beamlines.i24.jungfrau_commissioning.experiment_plans.rotation_scan_plan.calculate_motion_profile"
)
@patch(
    "mx_bluesky.beamlines.i24.jungfrau_commissioning.experiment_plans.rotation_scan_plan.set_up_beamline_for_rotation"
)
@patch(
    "mx_bluesky.beamlines.i24.jungfrau_commissioning.experiment_plans.rotation_scan_plan.fly_jungfrau"
)
async def test_rotation_scan_plan_in_re(
    mock_fly: MagicMock,
    mock_setup_beamline: MagicMock,
    mock_calc_motion_profile: MagicMock,
    mock_setup_zebra: MagicMock,
    mock_cleanup: MagicMock,
    run_engine: RunEngine,
    tmp_path,
    rotation_composite: RotationScanComposite,
):
    required_hardware_read_signals = [
        rotation_composite.dcm.energy_in_keV,
        rotation_composite.dcm.wavelength_in_a,
        rotation_composite.det_stage.z,
        rotation_composite.jungfrau._writer.file_path,
    ]

    rotation_composite.jungfrau._writer.final_path = (
        tmp_path  # Normally done during jf prepare
    )
    # Test correct functions are called, but don't test bluesky messages
    mock_zebra_arm = MagicMock(side_effect=lambda _: completed_status())
    rotation_composite.zebra.pc.arm.set = mock_zebra_arm
    params = get_good_single_rotation_params(tmp_path)
    mock_calc_motion_profile.return_value = calculate_motion_profile(params, 1, 1)
    run_engine(single_rotation_plan(rotation_composite, params))
    mock_setup_beamline.assert_called_once_with(
        rotation_composite, DEFAULT_DETECTOR_DISTANCE_MM, 0.1
    )
    mock_calc_motion_profile.assert_called_once_with(
        params, 1, await rotation_composite.gonio.omega.max_velocity.get_value()
    )
    mock_setup_zebra.assert_called_once()
    mock_zebra_arm.assert_called_once()
    mock_fly.assert_called_once()
    assert mock_fly.call_args_list[0][1]["read_hardware_after_prepare_plan"].args == (
        required_hardware_read_signals,
        PlanNameConstants.ROTATION_DEVICE_READ,
    )
    mock_cleanup.assert_called_once()


@patch(
    "mx_bluesky.beamlines.i24.jungfrau_commissioning.experiment_plans.rotation_scan_plan.set_up_beamline_for_rotation",
    new=MagicMock(),
)
@patch(
    "mx_bluesky.beamlines.i24.jungfrau_commissioning.experiment_plans.rotation_scan_plan.fly_jungfrau",
    new=MagicMock(),
)
def test_single_rotation_plan_in_simulator(
    sim_run_engine: RunEngineSimulator,
    rotation_composite: RotationScanComposite,
    tmp_path,
):
    params = get_good_single_rotation_params(tmp_path)
    set_mock_value(rotation_composite.hutch_shutter.status, ShutterState.OPEN)
    msgs = sim_run_engine.simulate_plan(
        single_rotation_plan(rotation_composite, params)
    )

    assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "open_run"
        and msg.run == "OUTER SINGLE ROTATION SCAN",
    )

    # Wait for rotation devices to be ready before reading metadata
    assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "wait"
        and msg.kwargs["group"] == PlanGroupCheckpointConstants.ROTATION_READY_FOR_DC,
    )

    # Set omega axis then wait for JF to complete
    assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set" and msg.obj == rotation_composite.gonio.omega,
    )
    assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "wait" and msg.kwargs["group"] == JF_COMPLETE_GROUP,
    )

    # Unstage JF and close run
    assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "unstage" and msg.obj == rotation_composite.jungfrau,
    )
    assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "close_run"
        and msg.run == PlanNameConstants.SINGLE_ROTATION_SCAN,
    )


@patch(
    "mx_bluesky.beamlines.i24.jungfrau_commissioning.experiment_plans.rotation_scan_plan.single_rotation_plan"
)
def test_rotation_plan_multiple_transmissions(
    mock_single_rotation: MagicMock,
    run_engine: RunEngine,
    tmp_path,
    rotation_composite: RotationScanComposite,
):
    desired_transmission_fracs = [0.2, 0.4, 0.6]
    params = get_good_multi_rotation_params(desired_transmission_fracs, tmp_path)
    set_mock_value(rotation_composite.hutch_shutter.status, ShutterState.OPEN)
    run_engine(rotation_scan_plan(rotation_composite, params))
    called_transmission_fracs = [
        mock_single_rotation.call_args_list[i].args[1].transmission_frac
        for i in range(mock_single_rotation.call_count)
    ]
    assert desired_transmission_fracs == called_transmission_fracs


async def test_set_up_beamline_for_rotation_success(
    rotation_composite: RotationScanComposite,
    run_engine: RunEngine,
):
    trans_frac = 0.1
    det_z = 200
    set_mock_value(rotation_composite.hutch_shutter.status, ShutterState.OPEN)
    run_engine(set_up_beamline_for_rotation(rotation_composite, det_z, trans_frac))

    assert await asyncio.gather(
        rotation_composite.aperture.position.get_value(),
        rotation_composite.beamstop.pos_select.get_value(),
        rotation_composite.det_stage.y.user_readback.get_value(),
        rotation_composite.backlight.backlight_position.pos_level.get_value(),
        rotation_composite.det_stage.z.user_readback.get_value(),
        rotation_composite.attenuator.actual_transmission.get_value(),
    ) == [
        AperturePositions.IN,
        BeamstopPositions.DATA_COLLECTION,
        JF_DET_STAGE_Y_POSITION_MM,
        BacklightPositions.OUT,
        det_z,
        trans_frac,
    ]


def test_set_up_beamline_for_rotation_error_on_closed_hutch(
    rotation_composite: RotationScanComposite,
    run_engine: RunEngine,
):
    trans_frac = 0.1
    det_z = 200
    set_mock_value(rotation_composite.hutch_shutter.status, ShutterState.CLOSED)
    with pytest.raises(HutchClosedError):
        run_engine(set_up_beamline_for_rotation(rotation_composite, det_z, trans_frac))


class FakeError(Exception): ...


@patch(
    "mx_bluesky.beamlines.i24.jungfrau_commissioning.experiment_plans.rotation_scan_plan.set_up_beamline_for_rotation",
)
def test_single_rotation_plan_uses_default_if_no_det_distance(
    mock_set_up_beamline: MagicMock,
    sim_run_engine: RunEngineSimulator,
    rotation_composite: RotationScanComposite,
    tmp_path,
):
    mock_set_up_beamline.side_effect = (
        FakeError,
    )  # Exit test early by inserting exception
    params = get_good_single_rotation_params(tmp_path)
    params.detector_distance_mm = None
    with pytest.raises(FakeError):
        sim_run_engine.simulate_plan(single_rotation_plan(rotation_composite, params))
    mock_set_up_beamline.assert_called_once_with(
        rotation_composite, DEFAULT_DETECTOR_DISTANCE_MM, params.transmission_frac
    )


@patch(
    "mx_bluesky.beamlines.i24.jungfrau_commissioning.experiment_plans.rotation_scan_plan.tidy_up_zebra_after_rotation_scan"
)
def test_cleanup_plan(
    mock_tidy_zebra: MagicMock,
    rotation_composite: RotationScanComposite,
    run_engine: RunEngine,
):
    rotation_composite.jungfrau.unstage = MagicMock(
        side_effect=lambda: completed_status()
    )
    run_engine(
        _cleanup_plan(
            rotation_composite.zebra,
            rotation_composite.jungfrau,
            rotation_composite.sample_shutter,
        )
    )
    mock_tidy_zebra.assert_called_once()
    rotation_composite.jungfrau.unstage.assert_called_once()



import os
from datetime import datetime
from pathlib import Path
from shutil import copy
from unittest.mock import patch

import bluesky.preprocessors as bpp
import h5py
import numpy as np
import pytest
from bluesky.run_engine import RunEngine
from h5py import Dataset, ExternalLink, Group

from mx_bluesky.common.experiment_plans.inner_plans.read_hardware import (
    standard_read_hardware_during_collection,
)
from mx_bluesky.common.external_interaction.nexus.write_nexus import NexusWriter
from mx_bluesky.common.parameters.rotation import (
    SingleRotationScan,
)
from mx_bluesky.common.utils.log import LOGGER
from mx_bluesky.hyperion.experiment_plans.rotation_scan_plan import (
    RotationScanComposite,
)
from mx_bluesky.hyperion.external_interaction.callbacks.rotation.nexus_callback import (
    RotationNexusFileCallback,
)
from mx_bluesky.hyperion.parameters.constants import CONST

from ....conftest import extract_metafile, raw_params_from_file

TEST_EXAMPLE_NEXUS_FILE = Path("ins_8_5.nxs")
TEST_EXAMPLE_NEXUS_METAFILE_PREFIX = "ins_8_5_meta"
TEST_DATA_DIRECTORY = Path("tests/test_data/nexus_files/rotation")
TEST_FILENAME = "rotation_scan_test_nexus"


@pytest.fixture
def test_params(tmp_path):
    param_dict = raw_params_from_file(
        "tests/test_data/parameter_json_files/good_test_rotation_scan_parameters.json",
        tmp_path,
    )
    param_dict["storage_directory"] = "tests/test_data"
    param_dict["file_name"] = f"{str(tmp_path)}/{TEST_FILENAME}"
    param_dict["scan_width_deg"] = 360.0
    param_dict["demand_energy_ev"] = 12700
    params = SingleRotationScan(**param_dict)
    params.x_start_um = 0
    params.y_start_um = 0
    params.z_start_um = 0
    params.exposure_time_s = 0.004
    return params


def fake_rotation_scan(
    parameters: SingleRotationScan,
    subscription: RotationNexusFileCallback,
    rotation_devices: RotationScanComposite,
):
    @bpp.subs_decorator(subscription)
    @bpp.set_run_key_decorator("rotation_scan_with_cleanup_and_subs")
    @bpp.run_decorator(  # attach experiment metadata to the start document
        md={
            "subplan_name": CONST.PLAN.ROTATION_OUTER,
            "mx_bluesky_parameters": parameters.model_dump_json(),
            "activate_callbacks": "RotationNexusFileCallback",
        }
    )
    def plan():
        yield from standard_read_hardware_during_collection(
            rotation_devices.aperture_scatterguard,
            rotation_devices.attenuator,
            rotation_devices.flux,
            rotation_devices.dcm,
            rotation_devices.eiger,
            rotation_devices.beamsize,
        )

    return plan()


def dectris_device_mapping(meta_filename: str):
    return {
        "entry": {
            "instrument": {
                "detector": {
                    "bit_depth_image": f"{meta_filename}//_dectris/bit_depth_image",
                    "bit_depth_readout": f"{meta_filename}//_dectris/bit_depth_image",
                    "detectorSpecific": {
                        "ntrigger": f"{meta_filename}///_dectris/ntrigger",
                        "software_version": f"{meta_filename}//_dectris/software_version",
                    },
                    "detector_readout_time": f"{meta_filename}//_dectris/detector_readout_time",
                    "flatfield_applied": f"{meta_filename}//_dectris/flatfield_correction_applied",
                    "photon_energy": f"{meta_filename}//_dectris/photon_energy",
                    "pixel_mask": f"{meta_filename}//mask",
                    "pixel_mask_applied": f"{meta_filename}//_dectris/pixel_mask_applied",
                    "threshold_energy": f"{meta_filename}//_dectris/threshold_energy",
                }
            }
        }
    }


def apply_metafile_mapping(exceptions: dict, mapping: dict):
    """Recursively populate the exceptions map with corresponding mapping entries"""
    for key in mapping.keys():
        mapping_value = mapping.get(key)
        if isinstance(mapping_value, dict):
            exceptions_child = exceptions.setdefault(key, {})
            apply_metafile_mapping(exceptions_child, mapping_value)
        else:
            exceptions[key] = mapping_value


@pytest.mark.timeout(2)
def test_rotation_scan_nexus_output_compared_to_existing_full_compare(
    test_params: SingleRotationScan,
    tmpdir,
    fake_create_rotation_devices: RotationScanComposite,
    run_engine: RunEngine,
):
    test_params.chi_start_deg = 0
    test_params.phi_start_deg = 0
    run_number = test_params.detector_params.run_number
    nexus_filename = f"{tmpdir}/{TEST_FILENAME}_{run_number}.nxs"
    master_filename = f"{tmpdir}/{TEST_FILENAME}_{run_number}_master.h5"
    meta_filename = f"{TEST_FILENAME}_{run_number}_meta.h5"

    fake_create_rotation_devices.eiger.bit_depth.sim_put(32)  # type: ignore

    run_engine(
        fake_rotation_scan(
            test_params, RotationNexusFileCallback(), fake_create_rotation_devices
        )
    )

    assert os.path.isfile(nexus_filename)
    assert os.path.isfile(master_filename)

    example_metafile_path = (
        f"{TEST_DATA_DIRECTORY}/{TEST_EXAMPLE_NEXUS_METAFILE_PREFIX}.h5.gz"
    )
    extract_metafile(
        example_metafile_path, f"{tmpdir}/{TEST_EXAMPLE_NEXUS_METAFILE_PREFIX}.h5"
    )
    example_nexus_path = f"{tmpdir}/{TEST_EXAMPLE_NEXUS_FILE}"
    copy(TEST_DATA_DIRECTORY / TEST_EXAMPLE_NEXUS_FILE, example_nexus_path)

    # Models expected differences to the GDA master nexus file
    # If a key is in _missing then it is not expected to be present
    # If a key is in _ignore then we do not compare it
    # If a key maps to a dict then we expect it to be a Group
    # Otherwise if a key is present we expect it to be a DataSet
    # If a key maps to a callable then we use that as the comparison function
    # Otherwise we compare the scalar or array value as appropriate
    exceptions = {
        "entry": {
            "_missing": {"end_time"},
            "data": {"_ignore": {"data", "omega"}},
            "instrument": {
                "_missing": {"transformations", "detector_z", "source"},
                "detector": {
                    "_missing": {
                        "detector_distance",
                        "serial_number",  # nexgen#236
                    },
                    "distance": 0.1,
                    "transformations": {
                        "detector_z": {"det_z": np.array([100])},
                        "det_z": np.array([100]),
                    },
                    "detector_z": {"det_z": np.array([100])},
                    "underload_value": 0,
                    "_ignore": {
                        "beam_center_x",
                        "beam_center_y",
                        "depends_on",
                        "saturation_value",
                        "sensor_material",
                    },
                    "detectorSpecific": {
                        "_missing": {"pixel_mask"},
                        "x_pixels": 4148,
                        "y_pixels": 4362,
                    },
                    "module": {"_ignore": {"module_offset"}},
                    "sensor_thickness": np.isclose,
                },
                "attenuator": {"attenuator_transmission": np.isclose},
                "beam": {"incident_wavelength": np.isclose},
                "name": b"DIAMOND BEAMLINE i03",
            },
            "sample": {
                "beam": {"incident_wavelength": np.isclose},
                "transformations": {
                    "_missing": {"omega_end"},
                    "_ignore": {"omega"},
                    "omega_increment_set": 0.1,
                    "omega_end": lambda a, b: np.all(np.isclose(a, b, atol=1e-03)),
                },
                "sample_omega": {
                    "_ignore": {"omega_end", "omega"},
                    "omega_increment_set": 0.1,
                },
                "sample_x": {"sam_x": np.isclose},
                "sample_y": {"sam_y": np.isclose},
                "sample_z": {"sam_z": np.isclose},
                "sample_chi": {"chi": np.isclose},
                "sample_phi": {"phi": np.isclose},
            },
            "_ignore": {
                "end_time_estimated",
                "start_time",
            },
            "source": {
                "name": b"Diamond Light Source",
                "type": b"Synchrotron X-ray Source",
            },
        }
    }

    with (
        h5py.File(example_nexus_path, "r") as example_nexus,
        h5py.File(nexus_filename, "r") as hyperion_nexus,
    ):
        apply_metafile_mapping(exceptions, dectris_device_mapping(meta_filename))
        _compare_actual_and_expected_nexus_output(
            hyperion_nexus, example_nexus, exceptions
        )


@pytest.mark.timeout(2)
def test_rotation_scan_nexus_output_compared_to_existing_file(
    test_params: SingleRotationScan,
    tmpdir,
    fake_create_rotation_devices: RotationScanComposite,
    run_engine: RunEngine,
):
    run_number = test_params.run_number or test_params.detector_params.run_number
    nexus_filename = f"{tmpdir}/{TEST_FILENAME}_{run_number}.nxs"
    master_filename = f"{tmpdir}/{TEST_FILENAME}_{run_number}_master.h5"

    fake_create_rotation_devices.eiger.bit_depth.sim_put(32)  # type: ignore

    run_engine(
        fake_rotation_scan(
            test_params, RotationNexusFileCallback(), fake_create_rotation_devices
        )
    )

    assert os.path.isfile(nexus_filename)
    assert os.path.isfile(master_filename)

    with (
        h5py.File(
            str(TEST_DATA_DIRECTORY / TEST_EXAMPLE_NEXUS_FILE), "r"
        ) as example_nexus,
        h5py.File(nexus_filename, "r") as hyperion_nexus,
    ):
        start_time: bytes = hyperion_nexus["/entry/start_time"][()]  # type: ignore
        end_time: bytes = hyperion_nexus["/entry/end_time_estimated"][()]  # type:ignore
        assert type(start_time) is type(example_nexus["/entry/start_time"][()])  # type: ignore
        assert type(end_time) is type(example_nexus["/entry/end_time"][()])  # type: ignore
        assert datetime.fromisoformat(start_time.decode("UTF-8"))
        assert datetime.fromisoformat(end_time.decode("UTF-8"))

        # we used to write the positions wrong...
        hyperion_omega: np.ndarray = np.array(
            hyperion_nexus["/entry/data/omega"][:]  # type: ignore
        )
        example_omega: np.ndarray = example_nexus["/entry/data/omega"][:]  # type: ignore
        assert np.allclose(hyperion_omega, example_omega)

        assert isinstance(
            hyperion_data := hyperion_nexus["/entry/data/data"], h5py.Dataset
        )
        example_data_shape = example_nexus["/entry/data/data"].shape  # type: ignore

        assert hyperion_data.dtype == "uint32"
        assert hyperion_data.shape == example_data_shape

        hyperion_instrument = hyperion_nexus["/entry/instrument"]
        example_instrument = example_nexus["/entry/instrument"]
        transmission = "attenuator/attenuator_transmission"
        wavelength = "beam/incident_wavelength"
        assert np.isclose(
            hyperion_instrument[transmission][()],  # type: ignore
            example_instrument[transmission][()],  # type: ignore
        )
        assert np.isclose(
            hyperion_instrument[wavelength][()],  # type: ignore
            example_instrument[wavelength][()],  # type: ignore
        )

        hyperion_sam_x = hyperion_nexus["/entry/sample/sample_x/sam_x"]
        example_sam_x = example_nexus["/entry/sample/sample_x/sam_x"]
        hyperion_sam_y = hyperion_nexus["/entry/sample/sample_y/sam_y"]
        example_sam_y = example_nexus["/entry/sample/sample_y/sam_y"]
        hyperion_sam_z = hyperion_nexus["/entry/sample/sample_z/sam_z"]
        example_sam_z = example_nexus["/entry/sample/sample_z/sam_z"]

        hyperion_sam_phi = hyperion_nexus["/entry/sample/sample_phi/phi"]
        example_sam_phi = example_nexus["/entry/sample/sample_phi/phi"]
        hyperion_sam_chi = hyperion_nexus["/entry/sample/sample_chi/chi"]
        example_sam_chi = example_nexus["/entry/sample/sample_chi/chi"]

        hyperion_sam_omega = hyperion_nexus["/entry/sample/sample_omega/omega"]
        example_sam_omega = example_nexus["/entry/sample/sample_omega/omega"]

        assert np.isclose(
            hyperion_sam_x[()],  # type: ignore
            example_sam_x[()],  # type: ignore
        )
        assert np.isclose(
            hyperion_sam_y[()],  # type: ignore
            example_sam_y[()],  # type: ignore
        )
        assert np.isclose(
            hyperion_sam_z[()],  # type: ignore
            example_sam_z[()],  # type: ignore
        )

        assert hyperion_sam_x.attrs.get("depends_on") == example_sam_x.attrs.get(
            "depends_on"
        )
        assert hyperion_sam_y.attrs.get("depends_on") == example_sam_y.attrs.get(
            "depends_on"
        )
        assert hyperion_sam_z.attrs.get("depends_on") == example_sam_z.attrs.get(
            "depends_on"
        )
        assert hyperion_sam_phi.attrs.get("depends_on") == example_sam_phi.attrs.get(
            "depends_on"
        )
        assert hyperion_sam_chi.attrs.get("depends_on") == example_sam_chi.attrs.get(
            "depends_on"
        )
        assert hyperion_sam_omega.attrs.get(
            "depends_on"
        ) == example_sam_omega.attrs.get("depends_on")


@pytest.mark.parametrize(
    "bit_depth,expected_type",
    [(8, np.uint8), (16, np.uint16), (32, np.uint32), (100, np.uint16)],
)
@patch("mx_bluesky.common.external_interaction.nexus.write_nexus.NXmxFileWriter")
def test_given_detector_bit_depth_changes_then_vds_datatype_as_expected(
    mock_nexus_writer,
    test_params: SingleRotationScan,
    fake_create_rotation_devices: RotationScanComposite,
    bit_depth,
    expected_type,
    run_engine: RunEngine,
):
    write_vds_mock = mock_nexus_writer.return_value.write_vds

    fake_create_rotation_devices.eiger.bit_depth.sim_put(bit_depth)  # type: ignore

    run_engine(
        fake_rotation_scan(
            test_params, RotationNexusFileCallback(), fake_create_rotation_devices
        )
    )

    for call in write_vds_mock.mock_calls:
        assert call.kwargs["vds_dtype"] == expected_type


def _compare_actual_and_expected_nexus_output(actual, expected, exceptions: dict):
    _compare_actual_and_expected([], actual, expected, exceptions)


def _compare_actual_and_expected(path: list[str], actual, expected, exceptions: dict):
    if expected is None:
        # The nexus file under test contains a node that isn't in the original GDA reference nexus file
        # but we may still expect something if the exception map contains it
        expected = {}

    path_str = "/".join(path)
    LOGGER.debug(f"Comparing {path_str}")
    keys_not_in_actual = (
        expected.keys() - actual.keys() - exceptions.get("_missing", set())
    )
    assert len(keys_not_in_actual) == 0, (
        f"Missing entries in group {path_str}, {keys_not_in_actual}"
    )

    keys_to_compare = actual.keys()
    keys_to_ignore = exceptions.get("_ignore", set())
    keys_not_in_expected = (
        keys_to_compare
        - expected.keys()
        - {k for k in exceptions.keys() if not k.startswith("_")}
        - keys_to_ignore
    )
    cmp = len(keys_not_in_expected) == 0
    keys_to_compare = sorted(keys_to_compare)
    assert cmp, f"Found unexpected entries in group {path_str}, {keys_not_in_expected}"
    for key in keys_to_compare:
        item_path = path + [key]
        actual_link = actual.get(key, getlink=True)
        item_path_str = "/" + "/".join(item_path)
        exception = exceptions.get(key, None)
        if isinstance(actual_link, ExternalLink):
            if exception:
                actual_link_path = f"{actual_link.filename}//{actual_link.path}"
                assert actual_link_path == exception, (
                    f"Actual and expected external links differ {actual_link_path}, {exception}"
                )
            else:
                LOGGER.debug(
                    f"Skipping external link {item_path_str} -> {actual_link.path}"
                )
            continue
        actual_class = actual.get(key, getclass=True, getlink=False)
        expected_class = (
            Group
            if isinstance(exception, dict)
            else (
                Dataset if exception is not None else expected.get(key, getclass=True)  # type: ignore
            )
        )
        actual_value = actual.get(key)
        expected_value = (
            expected.get(key)
            if (exception is None or isinstance(exception, dict) or callable(exception))
            else exception
        )
        if expected_class == Group:
            _compare_actual_and_expected(
                item_path, actual_value, expected.get(key), exceptions.get(key, {})
            )
        elif (expected_class == Dataset) and key not in keys_to_ignore:
            if isinstance(expected_value, Dataset):
                # Only check shape if we didn't override the expected value
                assert actual_value.shape == expected_value.shape, (
                    f"Actual and expected shapes differ for {item_path_str}: {actual_value.shape}, {expected_value.shape}"
                )
            else:
                assert hasattr(actual_value, "shape"), f"No shape for {item_path_str}"
                expected_shape = np.shape(expected_value)  # type: ignore
                assert actual_value.shape == expected_shape, (
                    f"{item_path_str} data shape not expected shape{actual_value.shape}, {expected_shape}"
                )
            if actual_value.shape == ():
                if callable(exception):
                    assert exceptions.get(key)(actual_value, expected_value)  # type: ignore
                elif np.isscalar(exception):
                    assert actual_value[()] == exception, (
                        f"{item_path_str} actual and expected did not match {actual_value[()]}, {exception}."
                    )
                else:
                    assert actual_class == expected_class, (
                        f"{item_path_str} Actual and expected class don't match {actual_class}, {expected_class}"
                    )
                    # fmt: off
                    assert (
                        actual_value[()] == expected_value[()]  # type: ignore
                    ), (
                        f"Actual and expected values differ for {item_path_str}: "
                        f"{actual_value[()]} != {expected_value[()]}"  # type: ignore
                    )
                    # fmt: on
            else:
                actual_value_str = np.array2string(actual_value, threshold=10)
                expected_value_str = np.array2string(expected_value, threshold=10)  # type: ignore
                if callable(exception):
                    assert exception(actual_value, expected_value), (
                        f"Actual and expected values differ for {item_path_str}: {actual_value_str} != {expected_value_str}, according to {exception}"
                    )
                else:
                    assert np.array_equal(
                        actual_value,
                        expected_value,  # type: ignore
                    ), (
                        f"Actual and expected values differ for {item_path_str}: {actual_value_str} != {expected_value_str}"
                    )


def test_override_parameters_override(test_params: SingleRotationScan):
    writer = NexusWriter(
        test_params, (1, 2, 3), {}, full_num_of_images=82367, meta_data_run_number=9852
    )
    assert writer.full_num_of_images != test_params.num_images
    assert writer.full_num_of_images == 82367
    assert writer.data_filename == f"{test_params.file_name}_9852"



from __future__ import annotations

from collections.abc import Callable, Generator
from typing import TYPE_CHECKING

from bluesky.callbacks import CallbackBase
from dodal.devices.zocalo import ZocaloStartInfo, ZocaloTrigger

from mx_bluesky.common.parameters.constants import (
    DocDescriptorNames,
)
from mx_bluesky.common.utils.exceptions import ISPyBDepositionNotMadeError
from mx_bluesky.common.utils.log import ISPYB_ZOCALO_CALLBACK_LOGGER

if TYPE_CHECKING:
    from event_model.documents import Event, EventDescriptor, RunStart, RunStop


ZocaloInfoGenerator = Generator[list[ZocaloStartInfo], dict, None]


class ZocaloCallback(CallbackBase):
    """Callback class to handle the triggering of Zocalo processing.
    Will start listening for collections when {triggering_plan} has been started.

    For every ispyb deposition that occurs inside this run the callback will send zocalo
    a run_start signal. Once the {triggering_plan} has ended the callback will send a
    run_end signal for all collections.

    Shouldn't be subscribed directly to the RunEngine, instead should be passed to the
    `emit` argument of an ISPyB callback which appends DCIDs to the relevant start doc.

    Args:
        triggering_plan: Name of the bluesky sub-plan inside of which we generate information
            to be submitted to zocalo; this is identified by the 'subplan_name' entry in the
            run start metadata.
        zocalo_environment: Name of the zocalo environment we use to connect to zocalo
        start_info_generator_factory: A factory method which returns a Generator,
            the generator is sent the ZOCALO_HW_READ event document and in return yields
            one or more ZocaloStartInfo which will each be submitted to zocalo as a job.
    """

    def __init__(
        self,
        triggering_plan: str,
        zocalo_environment: str,
        start_info_generator_factory: Callable[[], ZocaloInfoGenerator],
    ):
        super().__init__()
        self._info_generator_factory = start_info_generator_factory
        self.triggering_plan = triggering_plan
        self.zocalo_interactor = ZocaloTrigger(zocalo_environment)
        self._reset_state()

    def _reset_state(self):
        self.run_uid: str | None = None
        self.zocalo_info: list[ZocaloStartInfo] = []
        self._started_zocalo_collections: list[ZocaloStartInfo] = []
        self.descriptors: dict[str, EventDescriptor] = {}
        self._info_generator = self._info_generator_factory()
        # Prime the generator
        next(self._info_generator)

    def start(self, doc: RunStart):
        ISPYB_ZOCALO_CALLBACK_LOGGER.info("Zocalo handler received start document.")
        if self.triggering_plan and doc.get("subplan_name") == self.triggering_plan:
            self.run_uid = doc.get("uid")
        if self.run_uid:
            zocalo_infos = self._info_generator.send(doc)  # type: ignore
            self.zocalo_info.extend(zocalo_infos)

    def descriptor(self, doc: EventDescriptor):
        self.descriptors[doc["uid"]] = doc

    def event(self, doc: Event) -> Event:
        event_descriptor = self.descriptors[doc["descriptor"]]
        if event_descriptor.get("name") == DocDescriptorNames.ZOCALO_HW_READ:
            filename = doc["data"]["eiger_odin_file_writer_id"]
            for start_info in self.zocalo_info:
                start_info.filename = filename
                self.zocalo_interactor.run_start(start_info)
                self._started_zocalo_collections.append(start_info)
            self.zocalo_info = []
        return doc

    def stop(self, doc: RunStop):
        if doc.get("run_start") == self.run_uid:
            try:
                ISPYB_ZOCALO_CALLBACK_LOGGER.info(
                    f"Zocalo handler received stop document, for run {doc.get('run_start')}."
                )
                if not self._started_zocalo_collections:
                    raise ISPyBDepositionNotMadeError(
                        f"No ISPyB IDs received by the end of {self.triggering_plan=}"
                    )
                for info in self._started_zocalo_collections:
                    self.zocalo_interactor.run_end(info.ispyb_dcid)
            finally:
                self._reset_state()



import types
from functools import partial
from unittest.mock import MagicMock, call, patch

import bluesky.plan_stubs as bps
import bluesky.preprocessors as bpp
import numpy as np
import pytest
from bluesky.run_engine import RunEngine, RunEngineResult
from bluesky.simulators import assert_message_and_return_remaining
from bluesky.utils import FailedStatus, Msg
from dodal.beamlines import i03
from dodal.devices.detector.det_dim_constants import (
    EIGER_TYPE_EIGER2_X_16M,
)
from dodal.devices.fast_grid_scan import (
    ZebraFastGridScanThreeD,
    set_fast_grid_scan_params,
)
from dodal.devices.smargon import CombinedMove
from dodal.devices.synchrotron import SynchrotronMode
from dodal.devices.zocalo import ZocaloStartInfo
from ophyd.sim import NullStatus
from ophyd.status import Status
from ophyd_async.core import completed_status, set_mock_value

from mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan import (
    BeamlineSpecificFGSFeatures,
    FlyScanEssentialDevices,
    common_flyscan_xray_centre,
    kickoff_and_complete_gridscan,
    run_gridscan,
)
from mx_bluesky.common.experiment_plans.inner_plans.read_hardware import (
    read_hardware_plan,
)
from mx_bluesky.common.external_interaction.callbacks.common.zocalo_callback import (
    ZocaloCallback,
)
from mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_callback import (
    GridscanISPyBCallback,
    ispyb_activation_wrapper,
)
from mx_bluesky.common.external_interaction.callbacks.xray_centre.nexus_callback import (
    GridscanNexusFileCallback,
)
from mx_bluesky.common.external_interaction.ispyb.ispyb_store import (
    IspybIds,
)
from mx_bluesky.common.parameters.constants import DocDescriptorNames
from mx_bluesky.common.parameters.gridscan import SpecifiedThreeDGridScan
from mx_bluesky.common.utils.exceptions import (
    WarningError,
)
from tests.conftest import (
    RunEngineSimulator,
    create_dummy_scan_spec,
)

from ....conftest import TestData
from ...conftest import (
    ReWithSubs,
    modified_store_grid_scan_mock,
    run_generic_ispyb_handler_setup,
)


class CompleteError(Exception):
    # To avoid having to run through the entire plan during tests
    pass


def mock_plan():
    yield from bps.null()


@pytest.fixture
def run_engine_with_subs_snapshots_already_taken(run_engine_with_subs, test_event_data):
    run_engine, subscriptions = run_engine_with_subs
    ispyb_gridscan_callback = [
        sub for sub in subscriptions if isinstance(sub, GridscanISPyBCallback)
    ][0]
    ispyb_gridscan_callback.active = True
    ispyb_gridscan_callback.start(
        test_event_data.test_grid_detect_and_gridscan_start_document
    )  # type: ignore
    ispyb_gridscan_callback.start(test_event_data.test_gridscan_outer_start_document)  # type: ignore
    ispyb_gridscan_callback.descriptor(
        test_event_data.test_descriptor_document_oav_snapshot
    )
    ispyb_gridscan_callback.event(test_event_data.test_event_document_oav_snapshot_xy)
    ispyb_gridscan_callback.event(test_event_data.test_event_document_oav_snapshot_xz)
    return run_engine, subscriptions


@patch(
    "mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_callback.StoreInIspyb",
    modified_store_grid_scan_mock,
)
class TestFlyscanXrayCentrePlan:
    td: TestData = TestData()

    def test_eiger2_x_16_detector_specified(
        self,
        test_fgs_params: SpecifiedThreeDGridScan,
    ):
        assert (
            test_fgs_params.detector_params.detector_size_constants.det_type_string
            == EIGER_TYPE_EIGER2_X_16M
        )

    def test_when_run_gridscan_called_then_generator_returned(
        self,
    ):
        plan = run_gridscan(MagicMock(), MagicMock(), MagicMock())
        assert isinstance(plan, types.GeneratorType)

    def test_when_run_gridscan_called_ispyb_deposition_made_and_records_errors(
        self,
        run_engine: RunEngine,
        fake_fgs_composite: FlyScanEssentialDevices,
        test_fgs_params: SpecifiedThreeDGridScan,
        beamline_specific: BeamlineSpecificFGSFeatures,
    ):
        ispyb_callback = GridscanISPyBCallback(param_type=SpecifiedThreeDGridScan)
        run_engine.subscribe(ispyb_callback)

        error = None
        with patch.object(fake_fgs_composite.gonio.omega, "set") as mock_set:
            error = AssertionError("Test Exception")
            mock_set.side_effect = FailedStatus(error)
            with pytest.raises(FailedStatus):
                run_engine(
                    ispyb_activation_wrapper(
                        common_flyscan_xray_centre(
                            fake_fgs_composite, test_fgs_params, beamline_specific
                        ),
                        test_fgs_params,
                    ),
                )

        ispyb_callback.ispyb.end_deposition.assert_called_once_with(  # type: ignore
            IspybIds(data_collection_group_id=0, data_collection_ids=(0, 0)),
            "fail",
            "Test Exception",
        )

    @patch("bluesky.plan_stubs.abs_set", autospec=True)
    def test_results_passed_to_move_motors(
        self,
        bps_abs_set: MagicMock,
        test_fgs_params: SpecifiedThreeDGridScan,
        fake_fgs_composite: FlyScanEssentialDevices,
        run_engine: RunEngine,
    ):
        from mx_bluesky.common.device_setup_plans.manipulate_sample import move_x_y_z

        motor_position = (
            test_fgs_params.fast_gridscan_params.grid_position_to_motor_position(
                np.array([1, 2, 3])
            )
        )
        run_engine(move_x_y_z(fake_fgs_composite.gonio, *motor_position))
        bps_abs_set.assert_called_with(
            fake_fgs_composite.gonio,
            CombinedMove(x=motor_position[0], y=motor_position[1], z=motor_position[2]),
            group="move_x_y_z",
        )

    @patch(
        "mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan.run_gridscan",
    )
    @patch(
        "mx_bluesky.common.external_interaction.callbacks.common.zocalo_callback.ZocaloTrigger",
    )
    def test_individual_plans_triggered_once_and_only_once_in_composite_run(
        self,
        zoc_trigger: MagicMock,
        run_gridscan: MagicMock,
        run_engine_with_subs: ReWithSubs,
        fake_fgs_composite: FlyScanEssentialDevices,
        test_fgs_params: SpecifiedThreeDGridScan,
        beamline_specific: BeamlineSpecificFGSFeatures,
    ):
        run_engine, _ = run_engine_with_subs

        def wrapped_gridscan_and_move():
            yield from common_flyscan_xray_centre(
                fake_fgs_composite,
                test_fgs_params,
                beamline_specific,
            )

        run_engine(wrapped_gridscan_and_move())
        run_gridscan.assert_called_once()
        beamline_specific.setup_trigger_plan.assert_called_once()  # type: ignore
        beamline_specific.tidy_plan.assert_called_once()  # type: ignore

    @patch(
        "mx_bluesky.common.experiment_plans.inner_plans.do_fgs.check_topup_and_wait_if_necessary",
    )
    def test_waits_for_motion_program(
        self,
        check_topup_and_wait,
        run_engine: RunEngine,
        test_fgs_params: SpecifiedThreeDGridScan,
        fake_fgs_composite: FlyScanEssentialDevices,
    ):
        fake_fgs_composite.eiger.unstage = MagicMock(
            side_effect=lambda: completed_status()
        )
        fgs = i03.zebra_fast_grid_scan.build(connect_immediately=True, mock=True)
        fgs.KICKOFF_TIMEOUT = 0.1
        fgs.complete = MagicMock(side_effect=lambda: completed_status())
        set_mock_value(fgs.motion_program.running, 1)

        def test_plan():
            yield from kickoff_and_complete_gridscan(
                fgs,
                fake_fgs_composite.eiger,
                fake_fgs_composite.synchrotron,
                [
                    test_fgs_params.scan_points_first_grid,
                    test_fgs_params.scan_points_second_grid,
                ],
            )

        with pytest.raises(FailedStatus):
            run_engine(test_plan())
        fgs.KICKOFF_TIMEOUT = 1
        set_mock_value(fgs.motion_program.running, 0)
        set_mock_value(fgs.status, 1)
        res = run_engine(test_plan())

        assert isinstance(res, RunEngineResult)
        assert res.exit_status == "success"

    def test_if_gridscan_prepare_fails_with_invalid_grid_then_sample_exception_raised(
        self,
        run_engine: RunEngine,
        fake_fgs_composite: FlyScanEssentialDevices,
        beamline_specific: BeamlineSpecificFGSFeatures,
        test_fgs_params: SpecifiedThreeDGridScan,
    ):
        beamline_specific.set_flyscan_params_plan = partial(
            set_fast_grid_scan_params,
            beamline_specific.fgs_motors,
            test_fgs_params.fast_gridscan_params,
        )

        set_mock_value(beamline_specific.fgs_motors.device_scan_invalid, 1.0)  # type: ignore

        with pytest.raises(WarningError):
            run_engine(
                run_gridscan(fake_fgs_composite, test_fgs_params, beamline_specific)
            )

    @patch(
        "mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan.kickoff_and_complete_gridscan",
    )
    def test_if_gridscan_prepare_fails_with_other_exception_then_plan_re_raised(
        self,
        mock_kickoff_and_complete,
        run_engine: RunEngine,
        fake_fgs_composite: FlyScanEssentialDevices,
        beamline_specific: BeamlineSpecificFGSFeatures,
        test_fgs_params: SpecifiedThreeDGridScan,
    ):
        exception = FailedStatus()
        exception.__cause__ = Exception()

        beamline_specific.set_flyscan_params_plan = MagicMock(side_effect=exception)

        with pytest.raises(FailedStatus) as e:
            run_engine(
                run_gridscan(fake_fgs_composite, test_fgs_params, beamline_specific)
            )

        mock_kickoff_and_complete.assert_not_called()

        assert e.value == exception

    @patch(
        "mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan.bps.abs_set",
        autospec=True,
    )
    @patch(
        "mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan.bps.kickoff",
        autospec=True,
    )
    @patch(
        "mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan.bps.complete",
        autospec=True,
    )
    @patch(
        "mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan.bps.mv",
        autospec=True,
    )
    @patch(
        "mx_bluesky.common.external_interaction.nexus.write_nexus.NexusWriter",
        autospec=True,
        spec_set=True,
    )
    @patch(
        "mx_bluesky.common.experiment_plans.inner_plans.do_fgs.check_topup_and_wait_if_necessary",
        autospec=True,
    )
    def test_when_grid_scan_ran_then_eiger_disarmed_before_zocalo_end(
        self,
        mock_check_topup,
        nexuswriter,
        mock_mv,
        mock_complete,
        mock_kickoff,
        mock_abs_set,
        fake_fgs_composite: FlyScanEssentialDevices,
        test_fgs_params: SpecifiedThreeDGridScan,
        run_engine_with_subs_snapshots_already_taken: tuple[
            RunEngine,
            tuple[GridscanNexusFileCallback, GridscanISPyBCallback],
        ],
        beamline_specific: BeamlineSpecificFGSFeatures,
    ):
        test_fgs_params.x_steps = 9
        test_fgs_params.y_steps = 10
        test_fgs_params.z_steps = 12
        run_engine, (nexus_cb, ispyb_cb) = run_engine_with_subs_snapshots_already_taken
        # Put both mocks in a parent to easily capture order
        mock_parent = MagicMock()
        fake_fgs_composite.eiger.disarm_detector = mock_parent.disarm
        assert isinstance(ispyb_cb.emit_cb, ZocaloCallback)
        ispyb_cb.emit_cb.zocalo_interactor.run_end = mock_parent.run_end

        fake_fgs_composite.eiger.filewriters_finished = NullStatus()  # type: ignore
        fake_fgs_composite.eiger.odin.check_and_wait_for_odin_state = MagicMock(
            return_value=True
        )
        fake_fgs_composite.eiger.odin.file_writer.num_captured.sim_put(1200)  # type: ignore
        fake_fgs_composite.eiger.stage = MagicMock(
            return_value=Status(None, None, 0, True, True)
        )

        with patch(
            "mx_bluesky.common.external_interaction.callbacks.xray_centre.nexus_callback.NexusWriter.create_nexus_file",
            autospec=True,
        ):
            [run_engine.subscribe(cb) for cb in (nexus_cb, ispyb_cb)]
            run_engine(
                ispyb_activation_wrapper(
                    common_flyscan_xray_centre(
                        fake_fgs_composite, test_fgs_params, beamline_specific
                    ),
                    test_fgs_params,
                )
            )

        mock_parent.assert_has_calls(
            [call.disarm(), call.run_end(100), call.run_end(200)]
        )

    @patch(
        "mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan.bps.wait",
        autospec=True,
    )
    @patch(
        "mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan.bps.complete",
        autospec=True,
    )
    @patch(
        "mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan.bps.kickoff",
        autospec=True,
    )
    @patch(
        "mx_bluesky.common.experiment_plans.inner_plans.do_fgs.check_topup_and_wait_if_necessary",
        autospec=True,
    )
    def test_fgs_arms_eiger_without_grid_detect(
        self,
        mock_topup,
        mock_kickoff,
        mock_complete,
        mock_wait,
        fake_fgs_composite: FlyScanEssentialDevices,
        test_fgs_params: SpecifiedThreeDGridScan,
        run_engine: RunEngine,
        beamline_specific: BeamlineSpecificFGSFeatures,
    ):
        fake_fgs_composite.eiger.unstage = MagicMock(side_effect=completed_status)
        run_engine(run_gridscan(fake_fgs_composite, test_fgs_params, beamline_specific))
        fake_fgs_composite.eiger.stage.assert_called_once()  # type: ignore
        fake_fgs_composite.eiger.unstage.assert_called_once()

    @patch(
        "mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan.bps.kickoff",
        autospec=True,
    )
    @patch(
        "mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan.bps.wait",
        autospec=True,
    )
    @patch(
        "mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan.bps.complete",
        autospec=True,
    )
    @patch(
        "mx_bluesky.common.experiment_plans.inner_plans.do_fgs.check_topup_and_wait_if_necessary",
        autospec=True,
    )
    def test_when_grid_scan_fails_with_exception_then_detector_disarmed_and_correct_exception_returned(
        self,
        mock_topup,
        mock_complete,
        mock_wait,
        mock_kickoff,
        fake_fgs_composite: FlyScanEssentialDevices,
        test_fgs_params: SpecifiedThreeDGridScan,
        run_engine: RunEngine,
        beamline_specific: BeamlineSpecificFGSFeatures,
    ):
        beamline_specific.read_pre_flyscan_plan = partial(
            read_hardware_plan,
            [],
            DocDescriptorNames.HARDWARE_READ_DURING,
        )

        mock_complete.side_effect = CompleteError()

        fake_fgs_composite.eiger.stage = MagicMock(
            return_value=Status(None, None, 0, True, True)
        )

        fake_fgs_composite.eiger.filewriters_finished = NullStatus()

        fake_fgs_composite.eiger.odin.check_and_wait_for_odin_state = MagicMock()

        fake_fgs_composite.eiger.disarm_detector = MagicMock()
        fake_fgs_composite.eiger.disable_roi_mode = MagicMock()

        with pytest.raises(CompleteError):
            run_engine(
                bpp.run_wrapper(
                    run_gridscan(fake_fgs_composite, test_fgs_params, beamline_specific)
                )
            )

        fake_fgs_composite.eiger.disable_roi_mode.assert_called()
        fake_fgs_composite.eiger.disarm_detector.assert_called()

    @patch(
        "mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan.bps.kickoff",
        autospec=True,
    )
    @patch(
        "mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan.bps.complete",
        autospec=True,
    )
    @patch(
        "mx_bluesky.common.experiment_plans.inner_plans.do_fgs.check_topup_and_wait_if_necessary",
        autospec=True,
    )
    def test_kickoff_and_complete_gridscan_triggers_zocalo(
        self,
        mock_topup,
        mock_complete: MagicMock,
        mock_kickoff: MagicMock,
        run_engine_with_subs_snapshots_already_taken,
        fake_fgs_composite: FlyScanEssentialDevices,
        dummy_rotation_data_collection_group_info,
        zebra_fast_grid_scan: ZebraFastGridScanThreeD,
    ):
        id_1, id_2 = 100, 200

        run_engine, subs = run_engine_with_subs_snapshots_already_taken
        _, ispyb_cb = subs
        ispyb_cb.active = True
        ispyb_cb.ispyb = MagicMock()
        ispyb_cb.params = MagicMock()
        ispyb_cb.ispyb_ids.data_collection_ids = (id_1, id_2)
        ispyb_cb.data_collection_group_info = dummy_rotation_data_collection_group_info
        assert isinstance(ispyb_cb.emit_cb, ZocaloCallback)

        mock_zocalo_trigger = ispyb_cb.emit_cb.zocalo_interactor
        fake_fgs_composite.eiger.unstage = MagicMock(side_effect=completed_status)
        fake_fgs_composite.eiger.odin.file_writer.id.sim_put("test/filename")  # type: ignore

        x_steps, y_steps, z_steps = 10, 20, 30

        run_engine.subscribe(ispyb_cb)

        run_engine(
            kickoff_and_complete_gridscan(
                zebra_fast_grid_scan,
                fake_fgs_composite.eiger,
                fake_fgs_composite.synchrotron,
                scan_points=create_dummy_scan_spec(),
            )
        )

        expected_start_infos = [
            ZocaloStartInfo(id_1, "test/filename", 0, x_steps * y_steps, 0),
            ZocaloStartInfo(
                id_2, "test/filename", x_steps * y_steps, x_steps * z_steps, 1
            ),
        ]

        expected_start_calls = [
            call(expected_start_infos[0]),
            call(expected_start_infos[1]),
        ]

        assert mock_zocalo_trigger.run_start.call_count == 2  # type: ignore
        assert mock_zocalo_trigger.run_start.mock_calls == expected_start_calls  # type: ignore

        assert mock_zocalo_trigger.run_end.call_count == 2  # type: ignore
        assert mock_zocalo_trigger.run_end.mock_calls == [call(id_1), call(id_2)]  # type: ignore

    @patch(
        "mx_bluesky.common.experiment_plans.inner_plans.do_fgs.check_topup_and_wait_if_necessary",
        new=MagicMock(side_effect=lambda *_, **__: iter([Msg("check_topup")])),
    )
    def test_read_hardware_during_collection_occurs_after_eiger_arm(
        self,
        fake_fgs_composite: FlyScanEssentialDevices,
        test_fgs_params: SpecifiedThreeDGridScan,
        sim_run_engine: RunEngineSimulator,
        beamline_specific: BeamlineSpecificFGSFeatures,
    ):
        beamline_specific.read_during_collection_plan = partial(
            read_hardware_plan,
            [fake_fgs_composite.eiger.bit_depth],  # type: ignore
            DocDescriptorNames.HARDWARE_READ_DURING,
        )
        sim_run_engine.add_handler(
            "read",
            lambda msg: {"values": {"value": SynchrotronMode.USER}},
            "synchrotron-synchrotron_mode",
        )
        msgs = sim_run_engine.simulate_plan(
            run_gridscan(fake_fgs_composite, test_fgs_params, beamline_specific)
        )
        msgs = assert_message_and_return_remaining(
            msgs, lambda msg: msg.command == "stage" and msg.obj.name == "eiger"
        )
        msgs = assert_message_and_return_remaining(
            msgs,
            lambda msg: msg.command == "kickoff"
            and msg.obj == beamline_specific.fgs_motors,
        )
        msgs = assert_message_and_return_remaining(
            msgs, lambda msg: msg.command == "create"
        )
        msgs = assert_message_and_return_remaining(
            msgs,
            lambda msg: msg.command == "read" and msg.obj.name == "eiger_bit_depth",
        )
        msgs = assert_message_and_return_remaining(
            msgs, lambda msg: msg.command == "save"
        )

    @patch(
        "mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan.run_gridscan",
        autospec=True,
    )
    def test_when_gridscan_succeeds_and_results_fetched_ispyb_comment_appended_to(
        self,
        run_gridscan: MagicMock,
        run_engine_with_subs: tuple[
            RunEngine,
            tuple[GridscanNexusFileCallback, GridscanISPyBCallback],
        ],
        test_fgs_params: SpecifiedThreeDGridScan,
        fake_fgs_composite: FlyScanEssentialDevices,
        beamline_specific: BeamlineSpecificFGSFeatures,
    ):
        run_engine, (nexus_cb, ispyb_cb) = run_engine_with_subs

        def _wrapped_gridscan_and_move():
            run_generic_ispyb_handler_setup(ispyb_cb, test_fgs_params)
            yield from common_flyscan_xray_centre(
                fake_fgs_composite,
                test_fgs_params,
                beamline_specific,
            )

        run_engine(
            ispyb_activation_wrapper(_wrapped_gridscan_and_move(), test_fgs_params)
        )
        app_to_comment: MagicMock = ispyb_cb.ispyb.append_to_comment  # type:ignore
        app_to_comment.assert_called()
        append_aperture_call = app_to_comment.call_args_list[0].args[1]
        assert "Aperture:" in append_aperture_call



import gc
import os
from dataclasses import fields
from unittest.mock import patch
from weakref import WeakValueDictionary

import pytest
from blueapi.core import BlueskyContext
from bluesky.run_engine import get_bluesky_event_loop, set_bluesky_event_loop
from ophyd_async.core import Device
from ophyd_async.plan_stubs import ensure_connected

from mx_bluesky.common.utils.context import device_composite_from_context
from mx_bluesky.hyperion.baton_handler import _initialise_udc
from mx_bluesky.hyperion.experiment_plans.load_centre_collect_full_plan import (
    LoadCentreCollectComposite,
)
from mx_bluesky.hyperion.utils.context import setup_context

weak_ids_to_devices = WeakValueDictionary()


MAX_DEVICE_COUNT = 4


@pytest.fixture
def patch_setup_devices(request):
    dev_mode = request.param
    from mx_bluesky.hyperion.utils.context import setup_devices

    def patched_setup_devices(context: BlueskyContext, _: bool):
        setup_devices(context, dev_mode)

    with patch(
        "mx_bluesky.hyperion.baton_handler.setup_devices",
        side_effect=patched_setup_devices,
    ) as patched_func:
        yield patched_func


@pytest.fixture(autouse=True)
def restore_global_event_loop():
    """Constructing a RunEngine during the soak tests overwrites the global
    bluesky event loop, we must restore it to the global session fixture in order for
    subsequent tests to not be affected.
    """
    old_event_loop = get_bluesky_event_loop()
    try:
        yield
    finally:
        set_bluesky_event_loop(old_event_loop)


@pytest.fixture
def patch_ensure_connected():
    unpatched = ensure_connected

    def patched_func(*devices: Device, mock=False, timeout=1, force_reconnect=False):
        timeout = 1
        yield from unpatched(
            *devices, mock=mock, timeout=timeout, force_reconnect=force_reconnect
        )

    with patch(
        "blueapi.utils.connect_devices.ensure_connected", side_effect=patched_func
    ) as p:
        yield p


@pytest.mark.parametrize(
    "i, patch_setup_devices",
    [[i, True] for i in range(1, 101)],
    indirect=["patch_setup_devices"],
)
@pytest.mark.system_test
@patch.dict(os.environ, {"BEAMLINE": "i03"})
def test_udc_reloads_all_devices_soak_test_dev_mode(i: int, patch_setup_devices):
    reinitialise_beamline(True, i)


@pytest.mark.parametrize(
    "i, patch_setup_devices",
    [[i, False] for i in range(1, 101)],
    indirect=["patch_setup_devices"],
)
@patch.dict(os.environ, {"BEAMLINE": "i03"})
@patch("ophyd_async.plan_stubs._ensure_connected.DEFAULT_TIMEOUT", 1)
@pytest.mark.timeout(10)
def test_udc_reloads_all_devices_soak_test_real(
    i: int, patch_setup_devices, patch_ensure_connected
):
    """
    Deliberately not part of main system tests because this is SLOW and requires
    a beamline to connect to.
    """
    reinitialise_beamline(False, i)


def reinitialise_beamline(dev_mode: bool, i: int):
    """Reinitialise the beamline, which should cause all beamline devices
    to be reinstantiated, and then check for memory leaks (potentially caused
    if something holds a reference to the device).
    Since GC is in general implementation dependent and non-deterministic,
    it is not guaranteed that devices will be collected after only one iteration.
    Therefore this test is run a number of times in a loop and we check that the
    outstanding instances do not exceed some threshold value."""

    context = setup_context(dev_mode)
    devices_before_reset: LoadCentreCollectComposite = device_composite_from_context(
        context, LoadCentreCollectComposite
    )
    for f in fields(devices_before_reset):
        device = getattr(devices_before_reset, f.name)
        weak_ids_to_devices[id(device)] = device
    _initialise_udc(context, dev_mode)
    devices_after_reset: LoadCentreCollectComposite = device_composite_from_context(
        context, LoadCentreCollectComposite
    )
    for f in fields(devices_after_reset):
        device = getattr(devices_after_reset, f.name)
        weak_ids_to_devices[id(device)] = device
    for f in fields(devices_after_reset):
        device_after_reset = getattr(devices_after_reset, f.name)
        device_before_reset = getattr(devices_before_reset, f.name)
        assert device_before_reset is not device_after_reset, (
            f"{id(device_before_reset)} == {id(device_after_reset)}"
        )
    check_instances_are_garbage_collected(i)
    context.run_engine.loop.call_soon_threadsafe(context.run_engine.loop.stop)


def check_instances_are_garbage_collected(i: int):
    gc.collect()
    device_counts: dict[str, int] = {}
    for ref in weak_ids_to_devices.valuerefs():
        device = ref()
        if device is not None:
            device_counts[device.name] = device_counts.get(device.name, 0) + 1

    devices_by_count = sorted([(count, name) for name, count in device_counts.items()])
    print(
        f"Dictionary size is {len(weak_ids_to_devices)}, total live references is "
        f"{sum(device_and_count[0] for device_and_count in devices_by_count)}"
    )
    print(
        f"Max count device is {devices_by_count[-1]}, min count device is {devices_by_count[0]}"
    )
    for name, count in device_counts.items():
        max_count = min(MAX_DEVICE_COUNT, i * 2)
        assert count <= max_count, (
            f"Device count {name} exceeded max expected references {count}"
        )



from __future__ import annotations

from typing import Literal, cast

from pydantic import BaseModel, Field


class MultiXtalSelection(BaseModel):
    name: str
    ignore_xtal_not_found: bool = False


class TopNByMaxCountSelection(MultiXtalSelection):
    name: Literal["TopNByMaxCount"] = "TopNByMaxCount"  #  pyright: ignore [reportIncompatibleVariableOverride]
    n: int


class TopNByMaxCountForEachSampleSelection(MultiXtalSelection):
    name: Literal["TopNByMaxCountForEachSample"] = "TopNByMaxCountForEachSample"  #  pyright: ignore [reportIncompatibleVariableOverride]
    n: int


class WithCentreSelection(BaseModel):
    select_centres: TopNByMaxCountSelection | TopNByMaxCountForEachSampleSelection = (
        Field(discriminator="name", default=TopNByMaxCountSelection(n=1))
    )

    @property
    def selection_params(self) -> MultiXtalSelection:
        """A helper property because pydantic does not allow polymorphism with base classes
        # only type unions"""
        cast1 = cast(MultiXtalSelection, self.select_centres)
        return cast1



from functools import partial
from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest
from deepdiff.diff import DeepDiff
from dodal.devices.zebra.zebra import RotationDirection
from pydantic_extra_types.semantic_version import SemanticVersion

from mx_bluesky.common.parameters.components import (
    PARAMETER_VERSION,
    IspybExperimentType,
)
from mx_bluesky.hyperion._plan_runner_params import Wait
from mx_bluesky.hyperion.external_interaction.agamemnon import (
    _get_parameters_from_url,
    _get_pin_type_from_agamemnon_collect_parameters,
    _SinglePin,
    create_parameters_from_agamemnon,
    get_agamemnon_url,
)
from mx_bluesky.hyperion.parameters.load_centre_collect import LoadCentreCollect

EXPECTED_ROBOT_LOAD_AND_CENTRE_PARAMS = {
    "storage_directory": "/dls/tmp/data/year/cm00000-0/auto/test/xraycentring",
    "snapshot_directory": Path(
        "/dls/tmp/data/year/cm00000-0/auto/test/xraycentring/snapshots"
    ),
    "file_name": "test_xtal",
    "demand_energy_ev": 12700.045934258673,
    "tip_offset_um": 300,
    "grid_width_um": 600,
    "omega_start_deg": 0,
    "chi_start_deg": 0,
    "transmission_frac": 1.0,
}

EXPECTED_ROTATION_PARAMS = {
    "storage_directory": "/dls/tmp/data/year/cm00000-0/auto/test",
    "snapshot_directory": Path("/dls/tmp/data/year/cm00000-0/auto/test/snapshots"),
    "file_name": "test_xtal",
    "demand_energy_ev": 12700.045934258673,
    "exposure_time_s": 0.002,
    "snapshot_omegas_deg": [0, 90, 180, 270],
    "comment": "Complete_P1_sweep1 ",
    "transmission_frac": 0.5,
    "ispyb_experiment_type": IspybExperimentType.CHARACTERIZATION,
    "rotation_scans": [
        {
            "omega_start_deg": 0.0,
            "phi_start_deg": 0.0,
            "scan_width_deg": 360,
            "rotation_direction": RotationDirection.POSITIVE,
            "chi_start_deg": 0.0,
            "sample_id": 12345,
        }
    ],
}

EXPECTED_PARAMETERS = {
    "visit": "cm00000-0",
    "detector_distance_mm": 180.8,
    "sample_id": 12345,
    "sample_puck": 1,
    "sample_pin": 1,
    "parameter_model_version": SemanticVersion.validate_from_str(
        str(PARAMETER_VERSION)
    ),
    "select_centres": {
        "name": "TopNByMaxCount",
        "n": 1,
    },
    "robot_load_then_centre": EXPECTED_ROBOT_LOAD_AND_CENTRE_PARAMS,
    "multi_rotation_scan": EXPECTED_ROTATION_PARAMS,
}


@pytest.fixture()
def use_real_agamemnon(request):
    url = getattr(request, "param", None)

    def get_injected_url(url, beamline):
        return _get_parameters_from_url(url)

    with patch(
        "mx_bluesky.hyperion.external_interaction.agamemnon._get_next_instruction",
        side_effect=partial(get_injected_url, url),
    ) as patched_get_next_instruction:
        yield patched_get_next_instruction


@pytest.mark.requires(external="agamemnon")
def test_given_test_agamemnon_instruction_then_returns_none_loop_type():
    params = _get_parameters_from_url(get_agamemnon_url() + "/example/collect")
    loop_type = _get_pin_type_from_agamemnon_collect_parameters(params["collect"])
    assert loop_type == _SinglePin()


@pytest.mark.requires(external="agamemnon")
@pytest.mark.parametrize(
    "use_real_agamemnon", [f"{get_agamemnon_url()}/example/collect"], indirect=True
)
def test_given_test_agamemnon_instruction_then_load_centre_collect_parameters_populated(
    use_real_agamemnon: MagicMock,
):
    load_centre_collect = create_parameters_from_agamemnon()
    expected_parameter_model = [
        LoadCentreCollect.model_validate(EXPECTED_PARAMETERS),
        LoadCentreCollect.model_validate(EXPECTED_PARAMETERS),
    ]
    expected_parameter_model[1].robot_load_then_centre.chi_start_deg = 30.0
    expected_parameter_model[1].multi_rotation_scan.rotation_scans[
        0
    ].chi_start_deg = 30.0
    # Currently agamemnon example json has two different experiment types, although this is not
    # realistic
    expected_parameter_model[
        1
    ].multi_rotation_scan.ispyb_experiment_type = IspybExperimentType.OSC
    difference = DeepDiff(
        load_centre_collect,
        expected_parameter_model,
    )
    assert not difference


@pytest.mark.requires(external="agamemnon")
@pytest.mark.parametrize(
    "use_real_agamemnon", [f"{get_agamemnon_url()}/example/wait"], indirect=True
)
def test_create_parameters_from_agamemnon_decodes_wait_instruction(
    use_real_agamemnon: MagicMock,
):
    params = create_parameters_from_agamemnon()
    difference = DeepDiff(
        params,
        [
            Wait.model_validate(
                {
                    "duration_s": 10.0,
                }
            )
        ],
    )
    assert not difference



from pathlib import Path
from unittest.mock import patch

import pytest
from blueapi.config import ApplicationConfig
from blueapi.core import BlueskyContext
from blueapi.worker import Task
from bluesky import RunEngine

from .conftest import raw_params_from_file


@pytest.fixture
def bluesky_context(run_engine: RunEngine, use_beamline_i03):
    config = ApplicationConfig(
        **{  # type: ignore
            "env": {
                "sources": [
                    {
                        "kind": "deviceManager",
                        "module": "dodal.beamlines.i03",
                        "mock": True,
                    },
                    {
                        "kind": "planFunctions",
                        "module": "mx_bluesky.hyperion.blueapi.plans",
                    },
                ]
            }
        }
    )
    yield BlueskyContext(run_engine=run_engine, configuration=config)


def test_load_centre_collect(bluesky_context: BlueskyContext, tmp_path: Path):
    params = raw_params_from_file(
        "tests/test_data/parameter_json_files/external_load_centre_collect_params.json",
        tmp_path,
    )
    _call_blueapi_plan(
        bluesky_context,
        "load_centre_collect",
        "_load_centre_collect_full",
        {"parameters": params},
    )


def test_robot_unload(bluesky_context: BlueskyContext, tmp_path: Path):
    _call_blueapi_plan(
        bluesky_context, "robot_unload", "_robot_unload", {"visit": "cm12345-67"}
    )


def test_move_to_udc_default_state(bluesky_context: BlueskyContext):
    _call_blueapi_plan(
        bluesky_context, "move_to_udc_default_state", "_move_to_udc_default_state", {}
    )


def _call_blueapi_plan(
    bluesky_context: BlueskyContext,
    plan_name: str,
    internal_name: str,
    parameters: dict,
):
    with patch(
        f"mx_bluesky.hyperion.blueapi.in_process.{internal_name}",
        return_value=iter([]),
        create=False,
    ) as mock_plan:
        Task(name=plan_name, params=parameters).do_task(bluesky_context)

    mock_plan.assert_called_once()
